# Copyright (C) 2007,2012 Canonical Ltd
# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""An adapter between a Git Branch and a Bazaar Branch."""

import contextlib
from collections import defaultdict
from functools import partial
from io import BytesIO
from typing import Optional

from dulwich.config import ConfigFile as GitConfigFile
from dulwich.config import parse_submodules
from dulwich.objects import ZERO_SHA, NotCommitError
from dulwich.repo import check_ref_format

from .. import (
    branch,
    config,
    controldir,
    errors,
    lock,
    revision,
    trace,
    transport,
    urlutils,
)
from .. import repository as _mod_repository
from ..foreign import ForeignBranch
from ..revision import NULL_REVISION
from ..tag import InterTags, TagConflict, Tags, TagSelector, TagUpdates
from ..trace import is_quiet, mutter, warning
from .errors import NoPushSupport
from .mapping import decode_git_path
from .push import remote_divergence
from .refs import (
    branch_name_to_ref,
    is_tag,
    ref_to_branch_name,
    ref_to_tag_name,
    tag_name_to_ref,
)
from .unpeel_map import UnpeelMap
from .urls import bzr_url_to_git_url, git_url_to_bzr_url


def _update_tip(source, target, revid, overwrite=False):
    if not overwrite:
        last_rev = target.last_revision()
        graph = target.repository.get_graph(source.repository)
        if graph.is_ancestor(revid, last_rev):
            # target is ahead of revid
            return
        target.generate_revision_history(revid, last_rev, other_branch=source)
    else:
        target.generate_revision_history(revid)


def _calculate_revnos(branch):
    if branch._format.stores_revno():
        return True
    config = branch.get_config_stack()
    return config.get("calculate_revnos")


class GitPullResult(branch.PullResult):
    """Result of a pull from a Git branch."""

    def _lookup_revno(self, revid):
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        if not _calculate_revnos(self.target_branch):
            return None
        # Try in source branch first, it'll be faster
        with self.target_branch.lock_read():
            return self.target_branch.revision_id_to_revno(revid)

    @property
    def old_revno(self):
        return self._lookup_revno(self.old_revid)

    @property
    def new_revno(self):
        return self._lookup_revno(self.new_revid)


class InterTagsFromGitToRemoteGit(InterTags):
    @classmethod
    def is_compatible(klass, source, target):
        if not isinstance(source, GitTags):
            return False
        if not isinstance(target, GitTags):
            return False
        return not getattr(target.branch.repository, "_git", None) is not None

    def merge(
        self,
        overwrite: bool = False,
        ignore_master: bool = False,
        selector: Optional[TagSelector] = None,
    ) -> tuple[TagUpdates, set[TagConflict]]:
        if self.source.branch.repository.has_same_location(
            self.target.branch.repository
        ):
            return {}, set()
        updates = {}
        conflicts = []
        source_tag_refs = self.source.branch.get_tag_refs()
        ref_to_tag_map = {}

        def get_changed_refs(old_refs):
            ret = dict(old_refs)
            for ref_name, tag_name, peeled, unpeeled in source_tag_refs.iteritems():
                if selector and not selector(tag_name):
                    continue
                if old_refs.get(ref_name) == unpeeled:
                    pass
                elif overwrite or ref_name not in old_refs:
                    ret[ref_name] = unpeeled
                    updates[tag_name] = (
                        self.target.branch.repository.lookup_foreign_revision_id(peeled)
                    )
                    ref_to_tag_map[ref_name] = tag_name
                    self.target.branch._tag_refs = None
                else:
                    conflicts.append(
                        (
                            tag_name,
                            self.source.branch.repository.lookup_foreign_revision_id(
                                peeled
                            ),
                            self.target.branch.repository.lookup_foreign_revision_id(
                                old_refs[ref_name]
                            ),
                        )
                    )
            return ret

        result = self.target.branch.repository.controldir.send_pack(
            get_changed_refs, lambda have, want: []
        )
        if result is not None and not isinstance(result, dict):
            for ref, error in result.ref_status.items():
                if error:
                    warning("unable to update ref %s: %s", ref, error)
                    del updates[ref_to_tag_map[ref]]
        return updates, set(conflicts)


class InterTagsFromGitToLocalGit(InterTags):
    @classmethod
    def is_compatible(klass, source, target):
        if not isinstance(source, GitTags):
            return False
        if not isinstance(target, GitTags):
            return False
        return getattr(target.branch.repository, "_git", None) is not None

    def merge(self, overwrite=False, ignore_master=False, selector=None):
        if self.source.branch.repository.has_same_location(
            self.target.branch.repository
        ):
            return {}, []

        conflicts = []
        updates = {}
        source_tag_refs = self.source.branch.get_tag_refs()

        target_repo = self.target.branch.repository

        for ref_name, tag_name, peeled, unpeeled in source_tag_refs:
            if selector and not selector(tag_name):
                continue
            if target_repo._git.refs.get(ref_name) == unpeeled:
                pass
            elif overwrite or ref_name not in target_repo._git.refs:
                try:
                    updates[tag_name] = target_repo.lookup_foreign_revision_id(peeled)
                except KeyError:
                    trace.warning("%s does not point to a valid object", tag_name)
                    continue
                except NotCommitError:
                    trace.warning("%s points to a non-commit object", tag_name)
                    continue
                target_repo._git.refs[ref_name] = unpeeled or peeled
                self.target.branch._tag_refs = None
            else:
                try:
                    source_revid = (
                        self.source.branch.repository.lookup_foreign_revision_id(peeled)
                    )
                    target_revid = target_repo.lookup_foreign_revision_id(
                        target_repo._git.refs[ref_name]
                    )
                except KeyError:
                    trace.warning("%s does not point to a valid object", ref_name)
                    continue
                except NotCommitError:
                    trace.warning("%s points to a non-commit object", tag_name)
                    continue
                conflicts.append((tag_name, source_revid, target_revid))
        return updates, set(conflicts)


class InterTagsFromGitToNonGit(InterTags):
    @classmethod
    def is_compatible(klass, source: Tags, target: Tags):
        if not isinstance(source, GitTags):
            return False
        return not isinstance(target, GitTags)

    def merge(self, overwrite=False, ignore_master=False, selector=None):
        """See Tags.merge_to."""
        source_tag_refs = self.source.branch.get_tag_refs()
        if ignore_master:
            master = None
        else:
            master = self.target.branch.get_master_branch()
        with contextlib.ExitStack() as es:
            if master is not None:
                es.enter_context(master.lock_write())
            updates, conflicts = self._merge_to(
                self.target, source_tag_refs, overwrite=overwrite, selector=selector
            )
            if master is not None:
                extra_updates, extra_conflicts = self._merge_to(
                    master.tags,
                    overwrite=overwrite,
                    source_tag_refs=source_tag_refs,
                    ignore_master=ignore_master,
                    selector=selector,
                )
                updates.update(extra_updates)
                conflicts.update(extra_conflicts)
            return updates, conflicts

    def _merge_to(
        self,
        to_tags,
        source_tag_refs,
        overwrite=False,
        selector=None,
        ignore_master=False,
    ):
        unpeeled_map = defaultdict(set)
        conflicts = []
        updates = {}
        result = dict(to_tags.get_tag_dict())
        for _ref_name, tag_name, peeled, unpeeled in source_tag_refs:
            if selector and not selector(tag_name):
                continue
            if unpeeled is not None:
                unpeeled_map[peeled].add(unpeeled)
            try:
                bzr_revid = self.source.branch.lookup_foreign_revision_id(peeled)
            except NotCommitError:
                continue
            if result.get(tag_name) == bzr_revid:
                pass
            elif tag_name not in result or overwrite:
                result[tag_name] = bzr_revid
                updates[tag_name] = bzr_revid
            else:
                conflicts.append((tag_name, bzr_revid, result[tag_name]))
        to_tags._set_tag_dict(result)
        if len(unpeeled_map) > 0:
            map_file = UnpeelMap.from_repository(to_tags.branch.repository)
            map_file.update(unpeeled_map)
            map_file.save_in_repository(to_tags.branch.repository)
        return updates, set(conflicts)


InterTags.register_optimiser(InterTagsFromGitToRemoteGit)
InterTags.register_optimiser(InterTagsFromGitToLocalGit)
InterTags.register_optimiser(InterTagsFromGitToNonGit)


class GitTags(Tags):
    """Ref-based tag dictionary."""

    def __init__(self, branch):
        self.branch = branch
        self.repository = branch.repository

    def get_tag_dict(self):
        ret = {}
        for _ref_name, tag_name, peeled, _unpeeled in self.branch.get_tag_refs():
            try:
                bzr_revid = self.branch.lookup_foreign_revision_id(peeled)
            except NotCommitError:
                continue
            else:
                ret[tag_name] = bzr_revid
        return ret

    def lookup_tag(self, tag_name):
        """Return the referent string of a tag."""
        # TODO(jelmer): Replace with something more efficient for local tags.
        td = self.get_tag_dict()
        try:
            return td[tag_name]
        except KeyError:
            raise errors.NoSuchTag(tag_name)


class LocalGitTagDict(GitTags):
    """Dictionary with tags in a local repository."""

    def __init__(self, branch):
        super().__init__(branch)
        self.refs = self.repository.controldir._git.refs

    def _set_tag_dict(self, to_dict):
        extra = set(self.refs.allkeys())
        for k, revid in to_dict.items():
            name = tag_name_to_ref(k)
            if name in extra:
                extra.remove(name)
            try:
                self.set_tag(k, revid)
            except errors.GhostTagsNotSupported:
                pass
        for name in extra:
            if is_tag(name):
                del self.repository._git[name]

    def set_tag(self, name, revid):
        try:
            git_sha, mapping = self.branch.lookup_bzr_revision_id(revid)
        except errors.NoSuchRevision:
            raise errors.GhostTagsNotSupported(self)
        self.refs[tag_name_to_ref(name)] = git_sha
        self.branch._tag_refs = None

    def delete_tag(self, name):
        ref = tag_name_to_ref(name)
        if ref not in self.refs:
            raise errors.NoSuchTag(name)
        del self.refs[ref]
        self.branch._tag_refs = None


class GitBranchFormat(branch.BranchFormat):
    def network_name(self):
        return b"git"

    def supports_tags(self):
        return True

    def supports_leaving_lock(self):
        return False

    def supports_tags_referencing_ghosts(self):
        return False

    def tags_are_versioned(self):
        return False

    def get_foreign_tests_branch_factory(self):
        from .tests.test_branch import ForeignTestsBranchFactory

        return ForeignTestsBranchFactory()

    def make_tags(self, branch):
        try:
            return branch.tags
        except AttributeError:
            pass
        if getattr(branch.repository, "_git", None) is None:
            from .remote import RemoteGitTagDict

            return RemoteGitTagDict(branch)
        else:
            return LocalGitTagDict(branch)

    def initialize(
        self, a_controldir, name=None, repository=None, append_revisions_only=None
    ):
        raise NotImplementedError(self.initialize)

    def get_reference(self, controldir, name=None):
        return controldir.get_branch_reference(name=name)

    def set_reference(self, controldir, name, target):
        return controldir.set_branch_reference(target, name)

    def stores_revno(self):
        """True if this branch format store revision numbers."""
        return False

    supports_reference_locations = False


class LocalGitBranchFormat(GitBranchFormat):
    def get_format_description(self):
        return "Local Git Branch"

    @property
    def _matchingcontroldir(self):
        from .dir import LocalGitControlDirFormat

        return LocalGitControlDirFormat()

    def initialize(
        self, a_controldir, name=None, repository=None, append_revisions_only=None
    ):
        from .dir import LocalGitDir

        if not isinstance(a_controldir, LocalGitDir):
            raise errors.IncompatibleFormat(self, a_controldir._format)
        return a_controldir.create_branch(
            repository=repository,
            name=name,
            append_revisions_only=append_revisions_only,
        )


class GitBranch(ForeignBranch):
    """An adapter to git repositories for bzr Branch objects."""

    @property
    def control_transport(self):
        return self._control_transport

    @property
    def user_transport(self):
        return self._user_transport

    def __init__(self, controldir, repository, ref: bytes, format):
        self.repository = repository
        self._format = format
        self.controldir = controldir
        self._lock_mode = None
        self._lock_count = 0
        super().__init__(repository.get_mapping())
        if not isinstance(ref, bytes):
            raise TypeError("ref is invalid: {!r}".format(ref))
        self.ref = ref
        self._head = None
        self._user_transport = controldir.user_transport.clone(".")
        self._control_transport = controldir.control_transport.clone(".")
        self._tag_refs = None
        params: dict[str, str] = {}
        try:
            self.name = ref_to_branch_name(ref)
        except ValueError:
            self.name = None
            if self.ref is not None:
                params = {"ref": urlutils.escape(self.ref, safe="")}
        else:
            if self.name:
                params = {"branch": urlutils.escape(self.name, safe="")}
        for k, v in params.items():
            self._user_transport.set_segment_parameter(k, v)
            self._control_transport.set_segment_parameter(k, v)
        self.base = controldir.user_transport.base

    def _get_checkout_format(self, lightweight=False):
        """Return the most suitable metadir for a checkout of this branch.
        Weaves are used if this branch's repository uses weaves.
        """
        if lightweight:
            return controldir.format_registry.make_controldir("git")
        else:
            return controldir.format_registry.make_controldir("default")

    def set_stacked_on_url(self, url):
        raise branch.UnstackableBranchFormat(self._format, self.base)

    def get_child_submit_format(self):
        """Return the preferred format of submissions to this branch."""
        ret = self.get_config_stack().get("child_submit_format")
        if ret is not None:
            return ret
        return "git"

    def get_config(self):
        from .config import GitBranchConfig

        return GitBranchConfig(self)

    def get_config_stack(self):
        from .config import GitBranchStack

        return GitBranchStack(self)

    def _get_nick(self, local=False, possible_master_transports=None):
        """Find the nick name for this branch.

        :return: Branch nick
        """
        if getattr(self.repository, "_git", None):
            cs = self.repository._git.get_config_stack()
            try:
                return cs.get((b"branch", self.name.encode("utf-8")), b"nick").decode(
                    "utf-8"
                )
            except KeyError:
                pass
        return self.name or "HEAD"

    def _set_nick(self, nick):
        cf = self.repository._git.get_config()
        cf.set((b"branch", self.name.encode("utf-8")), b"nick", nick.encode("utf-8"))
        f = BytesIO()
        cf.write_to_file(f)
        self.repository._git._put_named_file("config", f.getvalue())

    nick = property(_get_nick, _set_nick)

    def __repr__(self):
        return "<{}({!r}, {!r})>".format(
            self.__class__.__name__, self.repository.base, self.name
        )

    def set_last_revision(self, revid):
        raise NotImplementedError(self.set_last_revision)

    def generate_revision_history(self, revid, last_rev=None, other_branch=None):
        if last_rev is not None:
            graph = self.repository.get_graph()
            if not graph.is_ancestor(last_rev, revid):
                # our previous tip is not merged into stop_revision
                raise errors.DivergedBranches(self, other_branch)

        self.set_last_revision(revid)

    def lock_write(self, token=None):
        if token is not None:
            raise errors.TokenLockingNotSupported(self)
        if self._lock_mode:
            if self._lock_mode == "r":
                raise errors.ReadOnlyError(self)
            self._lock_count += 1
        else:
            self._lock_ref()
            self._lock_mode = "w"
            self._lock_count = 1
        self.repository.lock_write()
        return lock.LogicalLockResult(self.unlock)

    def leave_lock_in_place(self):
        raise NotImplementedError(self.leave_lock_in_place)

    def dont_leave_lock_in_place(self):
        raise NotImplementedError(self.dont_leave_lock_in_place)

    def get_stacked_on_url(self):
        # Git doesn't do stacking (yet...)
        raise branch.UnstackableBranchFormat(self._format, self.base)

    def _get_push_origin(self, cs):
        """Get the name for the push origin.

        The exact behaviour is documented in the git-config(1) manpage.
        """
        try:
            return cs.get((b"branch", self.name.encode("utf-8")), b"pushRemote")
        except KeyError:
            try:
                return cs.get((b"branch",), b"remote")
            except KeyError:
                try:
                    return cs.get((b"branch", self.name.encode("utf-8")), b"remote")
                except KeyError:
                    return b"origin"

    def _get_origin(self, cs):
        try:
            return cs.get((b"branch", self.name.encode("utf-8")), b"remote")
        except KeyError:
            return b"origin"

    def _get_related_push_branch(self, cs):
        remote = self._get_push_origin(cs)
        try:
            location = cs.get((b"remote", remote), b"url")
        except KeyError:
            return None

        return git_url_to_bzr_url(location.decode("utf-8"), ref=self.ref)

    def _get_related_merge_branch(self, cs):
        remote = self._get_origin(cs)
        try:
            location = cs.get((b"remote", remote), b"url")
        except KeyError:
            return None

        try:
            ref = cs.get((b"branch", remote), b"merge")
        except KeyError:
            ref = b"HEAD"

        return git_url_to_bzr_url(location.decode("utf-8"), ref=ref)

    def _get_parent_location(self):
        """See Branch.get_parent()."""
        cs = self.repository._git.get_config_stack()
        return self._get_related_merge_branch(cs)

    def set_parent(self, location):
        cs = self.repository._git.get_config()
        remote = self._get_origin(cs)
        this_url = urlutils.strip_segment_parameters(self.user_url)
        target_url, branch, ref = bzr_url_to_git_url(location)
        location = urlutils.relative_url(this_url, target_url)
        cs.set((b"remote", remote), b"url", location)
        cs.set(
            (b"remote", remote), b"fetch", b"+refs/heads/*:refs/remotes/%s/*" % remote
        )
        if self.name:
            if branch:
                cs.set(
                    (b"branch", self.name.encode()),
                    b"merge",
                    branch_name_to_ref(branch),
                )
            elif ref:
                cs.set((b"branch", self.name.encode()), b"merge", ref)
            else:
                # TODO(jelmer): Maybe unset rather than setting to HEAD?
                cs.set((b"branch", self.name.encode()), b"merge", b"HEAD")
        self.repository._write_git_config(cs)

    def break_lock(self):
        raise NotImplementedError(self.break_lock)

    def lock_read(self):
        if self._lock_mode:
            if self._lock_mode not in ("r", "w"):
                raise ValueError(self._lock_mode)
            self._lock_count += 1
        else:
            self._lock_mode = "r"
            self._lock_count = 1
        self.repository.lock_read()
        return lock.LogicalLockResult(self.unlock)

    def peek_lock_mode(self):
        return self._lock_mode

    def is_locked(self):
        return self._lock_mode is not None

    def _lock_ref(self):
        pass

    def _unlock_ref(self):
        pass

    def unlock(self):
        """See Branch.unlock()."""
        if self._lock_count == 0:
            raise errors.LockNotHeld(self)
        try:
            self._lock_count -= 1
            if self._lock_count == 0:
                if self._lock_mode == "w":
                    self._unlock_ref()
                self._lock_mode = None
                self._clear_cached_state()
        finally:
            self.repository.unlock()

    def get_physical_lock_status(self):
        return False

    def last_revision(self):
        with self.lock_read():
            # perhaps should escape this ?
            if self.head is None:
                return revision.NULL_REVISION
            return self.lookup_foreign_revision_id(self.head)

    def _basic_push(
        self, target, overwrite=False, stop_revision=None, tag_selector=None
    ):
        return branch.InterBranch.get(self, target)._basic_push(
            overwrite, stop_revision, tag_selector=tag_selector
        )

    def lookup_foreign_revision_id(self, foreign_revid):
        try:
            return self.repository.lookup_foreign_revision_id(
                foreign_revid, self.mapping
            )
        except KeyError:
            # Let's try..
            return self.mapping.revision_id_foreign_to_bzr(foreign_revid)

    def lookup_bzr_revision_id(self, revid):
        return self.repository.lookup_bzr_revision_id(revid, mapping=self.mapping)

    def get_unshelver(self, tree):
        raise errors.StoringUncommittedNotSupported(self)

    def _clear_cached_state(self):
        super()._clear_cached_state()
        self._tag_refs = None

    def _iter_tag_refs(self, refs):
        """Iterate over the tag refs.

        :param refs: Refs dictionary (name -> git sha1)
        :return: iterator over (ref_name, tag_name, peeled_sha1, unpeeled_sha1)
        """
        raise NotImplementedError(self._iter_tag_refs)

    def get_tag_refs(self):
        with self.lock_read():
            if self._tag_refs is None:
                self._tag_refs = list(self._iter_tag_refs())
            return self._tag_refs

    def import_last_revision_info_and_tags(self, source, revno, revid, lossy=False):
        """Set the last revision info, importing from another repo if necessary.

        This is used by the bound branch code to upload a revision to
        the master branch first before updating the tip of the local branch.
        Revisions referenced by source's tags are also transferred.

        :param source: Source branch to optionally fetch from
        :param revno: Revision number of the new tip
        :param revid: Revision id of the new tip
        :param lossy: Whether to discard metadata that can not be
            natively represented
        :return: Tuple with the new revision number and revision id
            (should only be different from the arguments when lossy=True)
        """
        push_result = source.push(
            self, stop_revision=revid, lossy=lossy, _stop_revno=revno
        )
        return (push_result.new_revno, push_result.new_revid)

    def reconcile(self, thorough=True):
        """Make sure the data stored in this branch is consistent."""
        from ..reconcile import ReconcileResult

        # Nothing to do here
        return ReconcileResult()


class LocalGitBranch(GitBranch):
    """A local Git branch."""

    def __init__(self, controldir, repository, ref):
        super().__init__(controldir, repository, ref, LocalGitBranchFormat())

    def create_checkout(
        self,
        to_location,
        revision_id=None,
        lightweight=False,
        accelerator_tree=None,
        hardlink=False,
    ):
        t = transport.get_transport(to_location)
        t.ensure_base()
        format = self._get_checkout_format(lightweight=lightweight)
        checkout = format.initialize_on_transport(t)
        if lightweight:
            from_branch = checkout.set_branch_reference(target_branch=self)
        else:
            policy = checkout.determine_repository_policy()
            policy.acquire_repository()
            checkout_branch = checkout.create_branch()
            checkout_branch.bind(self)
            checkout_branch.pull(self, stop_revision=revision_id)
            from_branch = None
        return checkout.create_workingtree(
            revision_id, from_branch=from_branch, hardlink=hardlink
        )

    def _lock_ref(self):
        self._ref_lock = self.repository._git.refs.lock_ref(self.ref)

    def _unlock_ref(self):
        self._ref_lock.unlock()

    def break_lock(self):
        self.repository._git.refs.unlock_ref(self.ref)

    def _gen_revision_history(self):
        if self.head is None:
            return []
        last_revid = self.last_revision()
        graph = self.repository.get_graph()
        try:
            ret = list(
                graph.iter_lefthand_ancestry(last_revid, (revision.NULL_REVISION,))
            )
        except errors.RevisionNotPresent as e:
            raise errors.GhostRevisionsHaveNoRevno(last_revid, e.revision_id)
        ret.reverse()
        return ret

    def _get_head(self):
        try:
            return self.repository._git.refs[self.ref]
        except KeyError:
            return None

    def _read_last_revision_info(self):
        last_revid = self.last_revision()
        graph = self.repository.get_graph()
        try:
            revno = graph.find_distance_to_null(
                last_revid, [(revision.NULL_REVISION, 0)]
            )
        except errors.GhostRevisionsHaveNoRevno:
            revno = None
        return revno, last_revid

    def set_last_revision_info(self, revno, revision_id):
        self.set_last_revision(revision_id)
        self._last_revision_info_cache = revno, revision_id

    def set_last_revision(self, revid):
        if not revid or not isinstance(revid, bytes):
            raise errors.InvalidRevisionId(revision_id=revid, branch=self)
        if revid == NULL_REVISION:
            newhead = None
        else:
            (newhead, self.mapping) = self.repository.lookup_bzr_revision_id(revid)
            if self.mapping is None:
                raise AssertionError
        self._set_head(newhead)

    def _set_head(self, value):
        if value == ZERO_SHA:
            raise ValueError(value)
        self._head = value
        if value is None:
            del self.repository._git.refs[self.ref]
        else:
            self.repository._git.refs[self.ref] = self._head
        self._clear_cached_state()

    head = property(_get_head, _set_head)

    def get_push_location(self):
        """See Branch.get_push_location."""
        push_loc = self.get_config_stack().get("push_location")
        if push_loc is not None:
            return push_loc
        cs = self.repository._git.get_config_stack()
        return self._get_related_push_branch(cs)

    def set_push_location(self, location):
        """See Branch.set_push_location."""
        self.get_config().set_user_option(
            "push_location", location, store=config.STORE_LOCATION
        )

    def supports_tags(self):
        return True

    def store_uncommitted(self, creator):
        raise errors.StoringUncommittedNotSupported(self)

    def _iter_tag_refs(self):
        """Iterate over the tag refs.

        :param refs: Refs dictionary (name -> git sha1)
        :return: iterator over (ref_name, tag_name, peeled_sha1, unpeeled_sha1)
        """
        refs = self.repository.controldir.get_refs_container()
        for ref_name, unpeeled in refs.as_dict().items():
            try:
                tag_name = ref_to_tag_name(ref_name)
            except (ValueError, UnicodeDecodeError):
                continue
            peeled = refs.get_peeled(ref_name)
            if peeled is None:
                peeled = unpeeled
            if not isinstance(tag_name, str):
                raise TypeError(tag_name)
            yield (ref_name, tag_name, peeled, unpeeled)

    def create_memorytree(self):
        from .memorytree import GitMemoryTree

        return GitMemoryTree(self, self.repository._git.object_store, self.head)


def _quick_lookup_revno(local_branch, remote_branch, revid):
    if not isinstance(revid, bytes):
        raise TypeError(revid)
    # Try in source branch first, it'll be faster
    with local_branch.lock_read():
        if not _calculate_revnos(local_branch):
            return None
        try:
            return local_branch.revision_id_to_revno(revid)
        except errors.NoSuchRevision:
            graph = local_branch.repository.get_graph()
            try:
                return graph.find_distance_to_null(revid, [(revision.NULL_REVISION, 0)])
            except errors.GhostRevisionsHaveNoRevno:
                if not _calculate_revnos(remote_branch):
                    return None
                # FIXME: Check using graph.find_distance_to_null() ?
                with remote_branch.lock_read():
                    return remote_branch.revision_id_to_revno(revid)


class GitBranchPullResult(branch.PullResult):
    def __init__(self):
        super().__init__()
        self.new_git_head = None
        self._old_revno = None
        self._new_revno = None

    def report(self, to_file):
        if not is_quiet():
            if self.old_revid == self.new_revid:
                to_file.write("No revisions to pull.\n")
            elif self.new_git_head is not None:
                to_file.write(
                    f"Now on revision {self.new_revno} (git sha: {self.new_git_head}).\n"
                )
            else:
                to_file.write(f"Now on revision {self.new_revno}.\n")
        self._show_tag_conficts(to_file)

    def _lookup_revno(self, revid):
        return _quick_lookup_revno(self.target_branch, self.source_branch, revid)

    def _get_old_revno(self):
        if self._old_revno is not None:
            return self._old_revno
        return self._lookup_revno(self.old_revid)

    def _set_old_revno(self, revno):
        self._old_revno = revno

    old_revno = property(_get_old_revno, _set_old_revno)

    def _get_new_revno(self):
        if self._new_revno is not None:
            return self._new_revno
        return self._lookup_revno(self.new_revid)

    def _set_new_revno(self, revno):
        self._new_revno = revno

    new_revno = property(_get_new_revno, _set_new_revno)


class GitBranchPushResult(branch.BranchPushResult):
    def _lookup_revno(self, revid):
        return _quick_lookup_revno(self.source_branch, self.target_branch, revid)

    @property
    def old_revno(self):
        return self._lookup_revno(self.old_revid)

    @property
    def new_revno(self):
        new_original_revno = getattr(self, "new_original_revno", None)
        if new_original_revno:
            return new_original_revno
        if getattr(self, "new_original_revid", None) is not None:
            return self._lookup_revno(self.new_original_revid)
        return self._lookup_revno(self.new_revid)


class InterFromGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls from Git into bzr."""

    @staticmethod
    def _get_branch_formats_to_test():
        try:
            default_format = branch.format_registry.get_default()
        except AttributeError:
            default_format = branch.BranchFormat._default_format
        from .remote import RemoteGitBranchFormat

        return [
            (RemoteGitBranchFormat(), default_format),
            (LocalGitBranchFormat(), default_format),
        ]

    @classmethod
    def _get_interrepo(self, source, target):
        return _mod_repository.InterRepository.get(source.repository, target.repository)

    @classmethod
    def is_compatible(cls, source, target):
        if not isinstance(source, GitBranch):
            return False
        if isinstance(target, GitBranch):
            # InterLocalGitRemoteGitBranch or InterToGitBranch should be used
            return False
        if getattr(cls._get_interrepo(source, target), "fetch_objects", None) is None:  # noqa: SIM103
            # fetch_objects is necessary for this to work
            return False
        return True

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        self.fetch_objects(
            stop_revision, fetch_tags=fetch_tags, limit=limit, lossy=lossy
        )
        return _mod_repository.FetchResult()

    def fetch_objects(
        self, stop_revision, fetch_tags, limit=None, lossy=False, tag_selector=None
    ):
        interrepo = self._get_interrepo(self.source, self.target)
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get("branch.fetch_tags")

        def determine_wants(heads):
            if stop_revision is None:
                try:
                    head = heads[self.source.ref]
                except KeyError:
                    self._last_revid = revision.NULL_REVISION
                else:
                    self._last_revid = self.source.lookup_foreign_revision_id(head)
            else:
                self._last_revid = stop_revision
            real = interrepo.get_determine_wants_revids(
                [self._last_revid], include_tags=fetch_tags, tag_selector=tag_selector
            )
            return real(heads)

        pack_hint, head, refs = interrepo.fetch_objects(
            determine_wants, self.source.mapping, limit=limit, lossy=lossy
        )
        if pack_hint is not None and self.target.repository._format.pack_compresses:
            self.target.repository.pack(hint=pack_hint)
        return head, refs

    def _update_revisions(self, stop_revision=None, overwrite=False, tag_selector=None):
        head, refs = self.fetch_objects(
            stop_revision, fetch_tags=None, tag_selector=tag_selector
        )
        _update_tip(self.source, self.target, self._last_revid, overwrite)
        return head, refs

    def update_references(self, revid=None):
        if revid is None:
            revid = self.target.last_revision()
        tree = self.target.repository.revision_tree(revid)
        try:
            with tree.get_file(".gitmodules") as f:
                for path, url, _section in parse_submodules(GitConfigFile.from_file(f)):
                    self.target.set_reference_info(
                        tree.path2id(decode_git_path(path)),
                        url.decode("utf-8"),
                        decode_git_path(path),
                    )
        except transport.NoSuchFile:
            pass

    def _basic_pull(
        self,
        stop_revision,
        overwrite,
        run_hooks,
        _override_hook_target,
        _hook_master,
        tag_selector=None,
    ):
        if overwrite is True:
            overwrite = {"history", "tags"}
        elif not overwrite:
            overwrite = set()
        result = GitBranchPullResult()
        result.source_branch = self.source
        if _override_hook_target is None:
            result.target_branch = self.target
        else:
            result.target_branch = _override_hook_target
        with self.target.lock_write(), self.source.lock_read():
            # We assume that during 'pull' the target repository is closer than
            # the source one.
            (result.old_revno, result.old_revid) = self.target.last_revision_info()
            result.new_git_head, remote_refs = self._update_revisions(
                stop_revision,
                overwrite=("history" in overwrite),
                tag_selector=tag_selector,
            )
            tags_ret = self.source.tags.merge_to(
                self.target.tags, ("tags" in overwrite), ignore_master=True
            )
            if isinstance(tags_ret, tuple):
                result.tag_updates, result.tag_conflicts = tags_ret
            else:
                result.tag_conflicts = tags_ret
            (result.new_revno, result.new_revid) = self.target.last_revision_info()
            self.update_references(revid=result.new_revid)
            if _hook_master:
                result.master_branch = _hook_master
                result.local_branch = result.target_branch
            else:
                result.master_branch = result.target_branch
                result.local_branch = None
            if run_hooks:
                for hook in branch.Branch.hooks["post_pull"]:
                    hook(result)
            return result

    def pull(
        self,
        overwrite=False,
        stop_revision=None,
        possible_transports=None,
        _hook_master=None,
        run_hooks=True,
        _override_hook_target=None,
        local=False,
        tag_selector=None,
    ):
        """See Branch.pull.

        :param _hook_master: Private parameter - set the branch to
            be supplied as the master to pull hooks.
        :param run_hooks: Private parameter - if false, this branch
            is being called because it's the master of the primary branch,
            so it should not run its hooks.
        :param _override_hook_target: Private parameter - set the branch to be
            supplied as the target_branch to pull hooks.
        """
        # This type of branch can't be bound.
        bound_location = self.target.get_bound_location()
        if local and not bound_location:
            raise errors.LocalRequiresBoundBranch()
        source_is_master = False
        with contextlib.ExitStack() as es:
            es.enter_context(self.source.lock_read())
            if bound_location:
                # bound_location comes from a config file, some care has to be
                # taken to relate it to source.user_url
                normalized = urlutils.normalize_url(bound_location)
                try:
                    relpath = self.source.user_transport.relpath(normalized)
                    source_is_master = relpath == ""
                except (errors.PathNotChild, urlutils.InvalidURL):
                    source_is_master = False
            if not local and bound_location and not source_is_master:
                # not pulling from master, so we need to update master.
                master_branch = self.target.get_master_branch(possible_transports)
                es.enter_context(master_branch.lock_write())
                # pull from source into master.
                master_branch.pull(
                    self.source,
                    overwrite=overwrite,
                    stop_revision=stop_revision,
                    run_hooks=False,
                )
            else:
                master_branch = None
            return self._basic_pull(
                stop_revision,
                overwrite,
                run_hooks,
                _override_hook_target,
                _hook_master=master_branch,
                tag_selector=tag_selector,
            )

    def _basic_push(self, overwrite, stop_revision, tag_selector=None):
        if overwrite is True:
            overwrite = {"history", "tags"}
        elif not overwrite:
            overwrite = set()
        result = branch.BranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revno, result.old_revid = self.target.last_revision_info()
        result.new_git_head, remote_refs = self._update_revisions(
            stop_revision, overwrite=("history" in overwrite), tag_selector=tag_selector
        )
        tags_ret = self.source.tags.merge_to(
            self.target.tags,
            "tags" in overwrite,
            ignore_master=True,
            selector=tag_selector,
        )
        (result.tag_updates, result.tag_conflicts) = tags_ret
        result.new_revno, result.new_revid = self.target.last_revision_info()
        self.update_references(revid=result.new_revid)
        return result


class InterGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls between Git branches."""

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        raise NotImplementedError(self.fetch)


class InterLocalGitRemoteGitBranch(InterGitBranch):
    """InterBranch that copies from a local to a remote git branch."""

    @staticmethod
    def _get_branch_formats_to_test():
        from .remote import RemoteGitBranchFormat

        return [(LocalGitBranchFormat(), RemoteGitBranchFormat())]

    @classmethod
    def is_compatible(self, source, target):
        from .remote import RemoteGitBranch

        return isinstance(source, LocalGitBranch) and isinstance(
            target, RemoteGitBranch
        )

    def _basic_push(self, overwrite, stop_revision, tag_selector=None):
        from .remote import parse_git_error

        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        if stop_revision is None:
            stop_revision = self.source.last_revision()

        def get_changed_refs(old_refs):
            old_ref = old_refs.get(self.target.ref, None)
            if old_ref is None:
                result.old_revid = revision.NULL_REVISION
            else:
                result.old_revid = self.target.lookup_foreign_revision_id(old_ref)
            new_ref = self.source.repository.lookup_bzr_revision_id(stop_revision)[0]
            if not overwrite:
                if remote_divergence(
                    old_ref, new_ref, self.source.repository._git.object_store
                ):
                    raise errors.DivergedBranches(self.source, self.target)
            refs = {self.target.ref: new_ref}
            result.new_revid = stop_revision
            for name, sha in self.source.repository._git.refs.as_dict(
                b"refs/tags"
            ).items():
                if tag_selector and not tag_selector(name.decode("utf-8")):
                    continue
                if sha not in self.source.repository._git:
                    trace.mutter("Ignoring missing SHA: %s", sha)
                    continue
                refs[tag_name_to_ref(name)] = sha
            return refs

        dw_result = self.target.repository.send_pack(
            get_changed_refs, self.source.repository._git.generate_pack_data
        )
        if dw_result is not None and not isinstance(dw_result, dict):
            error = dw_result.ref_status.get(self.target.ref)
            if error:
                raise parse_git_error(self.target.user_url, error)
            for ref, error in dw_result.ref_status.items():
                if error:
                    trace.warning("unable to open ref %s: %s", ref, error)
        return result


class InterGitLocalGitBranch(InterGitBranch):
    """InterBranch that copies from a remote to a local git branch."""

    @staticmethod
    def _get_branch_formats_to_test():
        from .remote import RemoteGitBranchFormat

        return [
            (RemoteGitBranchFormat(), LocalGitBranchFormat()),
            (LocalGitBranchFormat(), LocalGitBranchFormat()),
        ]

    @classmethod
    def is_compatible(self, source, target):
        return isinstance(source, GitBranch) and isinstance(target, LocalGitBranch)

    def fetch(self, stop_revision=None, fetch_tags=None, limit=None, lossy=False):
        if lossy:
            raise errors.LossyPushToSameVCS(
                source_branch=self.source, target_branch=self.target
            )
        interrepo = _mod_repository.InterRepository.get(
            self.source.repository, self.target.repository
        )
        if stop_revision is None:
            stop_revision = self.source.last_revision()
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get("branch.fetch_tags")
        determine_wants = interrepo.get_determine_wants_revids(
            [stop_revision], include_tags=fetch_tags
        )
        interrepo.fetch_objects(determine_wants, limit=limit)
        return _mod_repository.FetchResult()

    def _basic_push(self, overwrite=False, stop_revision=None, tag_selector=None):
        if overwrite is True:
            overwrite = {"history", "tags"}
        elif not overwrite:
            overwrite = set()
        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.old_revid = self.target.last_revision()
        refs, stop_revision = self.update_refs(stop_revision)
        _update_tip(self.source, self.target, stop_revision, "history" in overwrite)
        tags_ret = self.source.tags.merge_to(
            self.target.tags, overwrite=("tags" in overwrite), selector=tag_selector
        )
        if isinstance(tags_ret, tuple):
            (result.tag_updates, result.tag_conflicts) = tags_ret
        else:
            result.tag_conflicts = tags_ret
        result.new_revid = self.target.last_revision()
        return result

    def update_refs(self, stop_revision=None):
        interrepo = _mod_repository.InterRepository.get(
            self.source.repository, self.target.repository
        )
        c = self.source.get_config_stack()
        fetch_tags = c.get("branch.fetch_tags")

        if stop_revision is None:
            result = interrepo.fetch(
                branches=[self.source.ref], include_tags=fetch_tags
            )
            try:
                head = result.refs[self.source.ref]
            except KeyError:
                stop_revision = revision.NULL_REVISION
            else:
                stop_revision = self.target.lookup_foreign_revision_id(head)
        else:
            result = interrepo.fetch(revision_id=stop_revision, include_tags=fetch_tags)
        return result.refs, stop_revision

    def pull(
        self,
        stop_revision=None,
        overwrite=False,
        possible_transports=None,
        run_hooks=True,
        local=False,
        tag_selector=None,
    ):
        # This type of branch can't be bound.
        if local:
            raise errors.LocalRequiresBoundBranch()
        if overwrite is True:
            overwrite = {"history", "tags"}
        elif not overwrite:
            overwrite = set()

        result = GitPullResult()
        result.source_branch = self.source
        result.target_branch = self.target
        with self.target.lock_write(), self.source.lock_read():
            result.old_revid = self.target.last_revision()
            refs, stop_revision = self.update_refs(stop_revision)
            _update_tip(self.source, self.target, stop_revision, "history" in overwrite)
            tags_ret = self.source.tags.merge_to(
                self.target.tags, overwrite=("tags" in overwrite), selector=tag_selector
            )
            if isinstance(tags_ret, tuple):
                (result.tag_updates, result.tag_conflicts) = tags_ret
            else:
                result.tag_conflicts = tags_ret
            result.new_revid = self.target.last_revision()
            result.local_branch = None
            result.master_branch = result.target_branch
            if run_hooks:
                for hook in branch.Branch.hooks["post_pull"]:
                    hook(result)
        return result


def _update_pure_git_refs(result, new_refs, overwrite, tag_selector, old_refs):
    mutter("updating refs. old refs: %r, new refs: %r", old_refs, new_refs)
    result.tag_updates = {}
    result.tag_conflicts = []
    ret = {}

    def ref_equals(refs, name, git_sha, revid):
        try:
            value = refs[name]
        except KeyError:
            return False
        if value[0] is not None and git_sha is not None:
            return value[0] == git_sha
        if value[1] is not None and revid is not None:
            return value[1] == revid

        # FIXME: If one side only has the git sha available and the other only
        # has the bzr revid, then this will cause us to show a tag as updated
        # that hasn't actually been updated.
        return False

    # FIXME: Check for diverged branches
    for ref, (git_sha, revid) in new_refs.items():
        if ref_equals(ret, ref, git_sha, revid):
            # Already up to date
            if git_sha is None:
                git_sha = old_refs[ref][0]
            if revid is None:
                revid = old_refs[ref][1]
            ret[ref] = new_refs[ref] = (git_sha, revid)
        elif ref not in ret or overwrite:
            try:
                tag_name = ref_to_tag_name(ref)
            except ValueError:
                pass
            else:
                if tag_selector and not tag_selector(tag_name):
                    continue
                result.tag_updates[tag_name] = revid
            ret[ref] = (git_sha, revid)
        else:
            # FIXME: Check diverged
            diverged = False
            if diverged:
                try:
                    name = ref_to_tag_name(ref)
                except ValueError:
                    pass
                else:
                    result.tag_conflicts.append((name, revid, ret[name][1]))
            else:
                ret[ref] = (git_sha, revid)
    return ret


class InterToGitBranch(branch.GenericInterBranch):
    """InterBranch implementation that pulls into a Git branch."""

    def __init__(self, source, target):
        super().__init__(source, target)
        self.interrepo = _mod_repository.InterRepository.get(
            source.repository, target.repository
        )

    @staticmethod
    def _get_branch_formats_to_test():
        try:
            default_format = branch.format_registry.get_default()
        except AttributeError:
            default_format = branch.BranchFormat._default_format
        from .remote import RemoteGitBranchFormat

        return [
            (default_format, LocalGitBranchFormat()),
            (default_format, RemoteGitBranchFormat()),
        ]

    @classmethod
    def is_compatible(self, source, target):
        return not isinstance(source, GitBranch) and isinstance(target, GitBranch)

    def _get_new_refs(self, stop_revision=None, fetch_tags=None, stop_revno=None):
        if not self.source.is_locked():
            raise errors.ObjectNotLocked(self.source)
        if stop_revision is None:
            (stop_revno, stop_revision) = self.source.last_revision_info()
        elif stop_revno is None:
            try:
                stop_revno = self.source.revision_id_to_revno(stop_revision)
            except errors.NoSuchRevision:
                stop_revno = None
        if not isinstance(stop_revision, bytes):
            raise TypeError(stop_revision)
        main_ref = self.target.ref
        refs = {main_ref: (None, stop_revision)}
        if fetch_tags is None:
            c = self.source.get_config_stack()
            fetch_tags = c.get("branch.fetch_tags")
        for name, revid in self.source.tags.get_tag_dict().items():
            if self.source.repository.has_revision(revid):
                ref = tag_name_to_ref(name)
                if not check_ref_format(ref):
                    warning("skipping tag with invalid characters %s (%s)", name, ref)
                    continue
                if fetch_tags:
                    # FIXME: Skip tags that are not in the ancestry
                    refs[ref] = (None, revid)
        return refs, main_ref, (stop_revno, stop_revision)

    def fetch(self, stop_revision=None, fetch_tags=None, lossy=False, limit=None):
        if stop_revision is None:
            stop_revision = self.source.last_revision()
        ret = []
        if fetch_tags:
            for _k, v in self.source.tags.get_tag_dict().items():
                ret.append((None, v))
        ret.append((None, stop_revision))
        if getattr(self.interrepo, "fetch_revs", None):
            try:
                revidmap = self.interrepo.fetch_revs(ret, lossy=lossy, limit=limit)
            except NoPushSupport:
                raise errors.NoRoundtrippingSupport(self.source, self.target)
            return _mod_repository.FetchResult(
                revidmap={
                    old_revid: new_revid
                    for (old_revid, (new_sha, new_revid)) in revidmap.items()
                }
            )
        else:

            def determine_wants(refs):
                wants = []
                for git_sha, revid in ret:
                    if git_sha is None:
                        git_sha, mapping = self.target.lookup_bzr_revision_id(revid)
                    wants.append(git_sha)
                return wants

            self.interrepo.fetch_objects(determine_wants, lossy=lossy, limit=limit)
            return _mod_repository.FetchResult()

    def pull(
        self,
        overwrite=False,
        stop_revision=None,
        local=False,
        possible_transports=None,
        run_hooks=True,
        _stop_revno=None,
        tag_selector=None,
    ):
        result = GitBranchPullResult()
        result.source_branch = self.source
        result.target_branch = self.target
        with self.source.lock_read(), self.target.lock_write():
            new_refs, main_ref, stop_revinfo = self._get_new_refs(
                stop_revision, stop_revno=_stop_revno
            )

            update_refs = partial(
                _update_pure_git_refs, result, new_refs, overwrite, tag_selector
            )
            try:
                result.revidmap, old_refs, new_refs = self.interrepo.fetch_refs(
                    update_refs, lossy=False
                )
            except NoPushSupport:
                raise errors.NoRoundtrippingSupport(self.source, self.target)
            (old_sha1, result.old_revid) = old_refs.get(
                main_ref, (ZERO_SHA, NULL_REVISION)
            )
            if result.old_revid is None:
                result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
            result.new_revid = new_refs[main_ref][1]
            result.local_branch = None
            result.master_branch = self.target
            if run_hooks:
                for hook in branch.Branch.hooks["post_pull"]:
                    hook(result)
        return result

    def push(
        self,
        overwrite=False,
        stop_revision=None,
        lossy=False,
        _override_hook_source_branch=None,
        _stop_revno=None,
        tag_selector=None,
    ):
        result = GitBranchPushResult()
        result.source_branch = self.source
        result.target_branch = self.target
        result.local_branch = None
        result.master_branch = result.target_branch
        with self.source.lock_read(), self.target.lock_write():
            new_refs, main_ref, stop_revinfo = self._get_new_refs(
                stop_revision, stop_revno=_stop_revno
            )

            update_refs = partial(
                _update_pure_git_refs, result, new_refs, overwrite, tag_selector
            )
            try:
                result.revidmap, old_refs, new_refs = self.interrepo.fetch_refs(
                    update_refs, lossy=lossy, overwrite=overwrite
                )
            except NoPushSupport:
                raise errors.NoRoundtrippingSupport(self.source, self.target)
            (old_sha1, result.old_revid) = old_refs.get(
                main_ref, (ZERO_SHA, NULL_REVISION)
            )
            if lossy or result.old_revid is None:
                result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
            result.new_revid = new_refs[main_ref][1]
            (result.new_original_revno, result.new_original_revid) = stop_revinfo
            for hook in branch.Branch.hooks["post_push"]:
                hook(result)
        return result


branch.InterBranch.register_optimiser(InterGitLocalGitBranch)
branch.InterBranch.register_optimiser(InterFromGitBranch)
branch.InterBranch.register_optimiser(InterToGitBranch)
branch.InterBranch.register_optimiser(InterLocalGitRemoteGitBranch)



# Copyright (C) 2018 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Export trees to tarballs, zipfiles, etc."""

from collections.abc import Iterator
from typing import cast

from .. import errors, registry


class ArchiveFormatInfo:
    def __init__(self, extensions):
        self.extensions = extensions


class ArchiveFormatRegistry(registry.Registry):
    """Registry of archive formats."""

    def __init__(self):
        self._extension_map = {}
        super().__init__()

    @property
    def extensions(self):
        return self._extension_map.keys()

    def register(self, key, factory, extensions, help=None):
        """Register an archive format."""
        registry.Registry.register(
            self, key, factory, help, ArchiveFormatInfo(extensions)
        )
        self._register_extensions(key, extensions)

    def register_lazy(self, key, module_name, member_name, extensions, help=None):
        registry.Registry.register_lazy(
            self, key, module_name, member_name, help, ArchiveFormatInfo(extensions)
        )
        self._register_extensions(key, extensions)

    def _register_extensions(self, name, extensions):
        for ext in extensions:
            self._extension_map[ext] = name

    def get_format_from_filename(self, filename):
        """Determine the archive format from an extension.

        :param filename: Filename to guess from
        :return: A format name, or None
        """
        for ext, format in self._extension_map.items():
            if filename.endswith(ext):
                return format
        else:
            return None


def create_archive(
    format, tree, name, root=None, subdir=None, force_mtime=None, recurse_nested=False
) -> Iterator[bytes]:
    try:
        archive_fn = format_registry.get(format)
    except KeyError as exc:
        raise errors.NoSuchExportFormat(format) from exc
    return cast(
        Iterator[bytes],
        archive_fn(
            tree,
            name,
            root=root,
            subdir=subdir,
            force_mtime=force_mtime,
            recurse_nested=recurse_nested,
        ),
    )


format_registry = ArchiveFormatRegistry()
format_registry.register_lazy(
    "tar",
    "breezy.archive.tar",
    "plain_tar_generator",
    [".tar"],
)
format_registry.register_lazy(
    "tgz", "breezy.archive.tar", "tgz_generator", [".tar.gz", ".tgz"]
)
format_registry.register_lazy(
    "tbz2", "breezy.archive.tar", "tbz_generator", [".tar.bz2", ".tbz2"]
)
format_registry.register_lazy(
    "tlzma", "breezy.archive.tar", "tar_lzma_generator", [".tar.lzma"]
)
format_registry.register_lazy(
    "txz", "breezy.archive.tar", "tar_xz_generator", [".tar.xz"]
)
format_registry.register_lazy(
    "zip", "breezy.archive.zip", "zip_archive_generator", [".zip"]
)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re

from .. import errors, gpg, mail_client, merge_directive, tests, trace

OUTPUT1 = b"""# Bazaar merge directive format 1
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
#\x20
booga"""

OUTPUT1_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.90)
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# base_revision_id: null:
#\x20
# Begin bundle
booga"""

OUTPUT2 = b"""# Bazaar merge directive format 1
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
#\x20
booga"""

OUTPUT2_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.90)
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
# base_revision_id: null:
#\x20
# Begin patch
booga"""

INPUT1 = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 1\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
#\x20
booga""".splitlines(True)


INPUT1_2 = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 2 (Bazaar 0.90)\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# base_revision_id: null:
# message: Hi mom!
#\x20
# Begin patch
booga""".splitlines(True)


INPUT1_2_OLD = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 2 (Bazaar 0.19)\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# base_revision_id: null:
# message: Hi mom!
#\x20
# Begin patch
booga""".splitlines(True)


OLD_DIRECTIVE_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.19)
# revision_id: abentley@panoramicfeedback.com-20070807234458-\
#   nzhkoyza56lan7z5
# target_branch: http://panoramicfeedback.com/opensource/bzr/repo\
#   /bzr.ab
# testament_sha1: d825a5cdb267a90ec2ba86b00895f3d8a9bed6bf
# timestamp: 2007-08-10 16:15:02 -0400
# source_branch: http://panoramicfeedback.com/opensource/bzr/repo\
#   /bzr.ab
# base_revision_id: abentley@panoramicfeedback.com-20070731163346-\
#   623xwcycwij91xen
#
""".splitlines(True)


class TestMergeDirective:
    def test_merge_source(self):
        time = 500000.0
        timezone = 5 * 3600
        self.assertRaises(
            errors.NoMergeSource,
            self.make_merge_directive,
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
        )
        self.assertRaises(
            errors.NoMergeSource,
            self.make_merge_directive,
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            patch_type="diff",
        )
        self.make_merge_directive(
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            source_branch="http://example.org",
        )
        md = self.make_merge_directive(
            b"null:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            patch=b"blah",
            patch_type="bundle",
        )
        self.assertIs(None, md.source_branch)
        md2 = self.make_merge_directive(
            b"null:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            patch=b"blah",
            patch_type="bundle",
            source_branch="bar",
        )
        self.assertEqual("bar", md2.source_branch)

    def test_serialization(self):
        time = 453
        timezone = 120
        md = self.make_merge_directive(
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            patch=b"booga",
            patch_type="bundle",
        )
        self.assertEqualDiff(self.OUTPUT1, b"".join(md.to_lines()))
        md = self.make_merge_directive(
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            source_branch="http://example.org",
            patch=b"booga",
            patch_type="diff",
            message="Hi mom!",
        )
        self.assertEqualDiff(self.OUTPUT2, b"".join(md.to_lines()))

    def test_deserialize_junk(self):
        self.assertRaises(
            errors.NotAMergeDirective,
            merge_directive.MergeDirective.from_lines,
            [b"lala"],
        )

    def test_deserialize_empty(self):
        self.assertRaises(
            errors.NotAMergeDirective, merge_directive.MergeDirective.from_lines, []
        )

    def test_deserialize_leading_junk(self):
        md = merge_directive.MergeDirective.from_lines(self.INPUT1)
        self.assertEqual(b"example:", md.revision_id)
        self.assertEqual(b"sha", md.testament_sha1)
        self.assertEqual("http://example.com", md.target_branch)
        self.assertEqual("http://example.org", md.source_branch)
        self.assertEqual(453, md.time)
        self.assertEqual(120, md.timezone)
        self.assertEqual(b"booga", md.patch)
        self.assertEqual("diff", md.patch_type)
        self.assertEqual("Hi mom!", md.message)

    def test_roundtrip(self):
        time = 500000
        timezone = 7.5 * 3600
        md = self.make_merge_directive(
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            source_branch="http://example.org",
            patch=b"booga",
            patch_type="diff",
        )
        md2 = merge_directive.MergeDirective.from_lines(md.to_lines())
        self.assertEqual(b"example:", md2.revision_id)
        self.assertIsInstance(md2.revision_id, bytes)
        self.assertEqual(b"sha", md2.testament_sha1)
        self.assertEqual("http://example.com", md2.target_branch)
        self.assertEqual("http://example.org", md2.source_branch)
        self.assertEqual(time, md2.time)
        self.assertEqual(timezone, md2.timezone)
        self.assertEqual("diff", md2.patch_type)
        self.assertEqual(b"booga", md2.patch)
        self.assertEqual(None, md2.message)
        self.set_bundle(md, b"# Bazaar revision bundle v0.9\n#\n")
        md.message = "Hi mom!"
        lines = md.to_lines()
        md3 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(b"# Bazaar revision bundle v0.9\n#\n", md3.bundle)
        self.assertEqual("bundle", md3.patch_type)
        self.assertContainsRe(md3.to_lines()[0], b"^# Bazaar merge directive format ")
        self.assertEqual("Hi mom!", md3.message)
        md3.clear_payload()
        self.assertIs(None, md3.get_raw_bundle())
        md4 = merge_directive.MergeDirective.from_lines(md3.to_lines())
        self.assertIs(None, md4.patch_type)


class TestMergeDirective1(tests.TestCase, TestMergeDirective):
    """Test merge directive format 1."""

    INPUT1 = INPUT1

    OUTPUT1 = OUTPUT1

    OUTPUT2 = OUTPUT2

    def make_merge_directive(
        self,
        revision_id,
        testament_sha1,
        time,
        timezone,
        target_branch,
        patch=None,
        patch_type=None,
        source_branch=None,
        message=None,
    ):
        return merge_directive.MergeDirective(
            revision_id,
            testament_sha1,
            time,
            timezone,
            target_branch,
            patch,
            patch_type,
            source_branch,
            message,
        )

    @staticmethod
    def set_bundle(md, value):
        md.patch = value

    def test_require_patch(self):
        time = 500.0
        timezone = 120
        self.assertRaises(
            errors.PatchMissing,
            merge_directive.MergeDirective,
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            patch_type="bundle",
        )
        md = merge_directive.MergeDirective(
            b"example:",
            b"sha1",
            time,
            timezone,
            "http://example.com",
            source_branch="http://example.org",
            patch=b"",
            patch_type="diff",
        )
        self.assertEqual(md.patch, b"")


class TestMergeDirective2(tests.TestCase, TestMergeDirective):
    """Test merge directive format 2."""

    INPUT1 = INPUT1_2

    OUTPUT1 = OUTPUT1_2

    OUTPUT2 = OUTPUT2_2

    def make_merge_directive(
        self,
        revision_id,
        testament_sha1,
        time,
        timezone,
        target_branch,
        patch=None,
        patch_type=None,
        source_branch=None,
        message=None,
        base_revision_id=b"null:",
    ):
        if patch_type == "bundle":
            bundle = patch
            patch = None
        else:
            bundle = None
        return merge_directive.MergeDirective2(
            revision_id,
            testament_sha1,
            time,
            timezone,
            target_branch,
            patch,
            source_branch,
            message,
            bundle,
            base_revision_id,
        )

    @staticmethod
    def set_bundle(md, value):
        md.bundle = value


EMAIL1 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 1
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
"""


EMAIL1_2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 2 \\(Bazaar 0.90\\)
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
"""


EMAIL2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a with special message
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 1
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
# message: Commit of rev2a with special message
"""

EMAIL2_2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a with special message
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 2 \\(Bazaar 0.90\\)
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
# message: Commit of rev2a with special message
"""


class TestMergeDirectiveBranch:
    def make_trees(self):
        tree_a = self.make_branch_and_tree("tree_a")
        tree_a.branch.get_config_stack().set(
            "email", "J. Random Hacker <jrandom@example.com>"
        )
        self.build_tree_contents(
            [
                ("tree_a/file", b"content_a\ncontent_b\n"),
                ("tree_a/file_2", b"content_x\rcontent_y\r"),
            ]
        )
        tree_a.add(["file", "file_2"])
        tree_a.commit("message", rev_id=b"rev1")
        tree_b = tree_a.controldir.sprout("tree_b").open_workingtree()
        branch_c = tree_a.controldir.sprout("branch_c").open_branch()
        tree_b.commit("message", rev_id=b"rev2b")
        self.build_tree_contents(
            [
                ("tree_a/file", b"content_a\ncontent_c \n"),
                ("tree_a/file_2", b"content_x\rcontent_z\r"),
            ]
        )
        tree_a.commit("Commit of rev2a", rev_id=b"rev2a")
        return tree_a, tree_b, branch_c

    def test_empty_target(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_d = self.make_branch_and_tree("tree_d")
        self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            120,
            tree_d.branch.base,
            patch_type="diff",
            public_branch=tree_a.branch.base,
        )

    def test_disk_name(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_a.branch.nick = "fancy <name>"
        md = self.from_objects(
            tree_a.branch.repository, b"rev2a", 500, 120, tree_b.branch.base
        )
        self.assertEqual("fancy-name-2", md.get_disk_name(tree_a.branch))

    def test_disk_name_old_revno(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_a.branch.nick = "fancy-name"
        md = self.from_objects(
            tree_a.branch.repository, b"rev1", 500, 120, tree_b.branch.base
        )
        self.assertEqual("fancy-name-1", md.get_disk_name(tree_a.branch))

    def test_generate_patch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md2 = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            120,
            tree_b.branch.base,
            patch_type="diff",
            public_branch=tree_a.branch.base,
        )
        self.assertNotContainsRe(md2.patch, b"Bazaar revision bundle")
        self.assertContainsRe(md2.patch, b"\\+content_c")
        self.assertNotContainsRe(md2.patch, b"\\+\\+\\+ b/")
        self.assertContainsRe(md2.patch, b"\\+\\+\\+ file")

    def test_public_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        self.assertRaises(
            errors.PublicBranchOutOfDate,
            self.from_objects,
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            public_branch=branch_c.base,
            patch_type="diff",
        )
        self.assertRaises(
            errors.PublicBranchOutOfDate,
            self.from_objects,
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            public_branch=branch_c.base,
            patch_type=None,
        )
        # public branch is not checked if patch format is bundle.
        md1 = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            public_branch=branch_c.base,
        )
        # public branch is provided with a bundle, despite possibly being out
        # of date, because it's not required if a bundle is present.
        self.assertEqual(md1.source_branch, branch_c.base)
        # Once we update the public branch, we can generate a diff.
        branch_c.pull(tree_a.branch)
        self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            patch_type=None,
            public_branch=branch_c.base,
        )

    def test_use_public_submit_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            patch_type=None,
            public_branch=branch_c.base,
        )
        self.assertEqual(md.target_branch, tree_b.branch.base)
        tree_b.branch.set_public_branch("http://example.com")
        md2 = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            144,
            tree_b.branch.base,
            patch_type=None,
            public_branch=branch_c.base,
        )
        self.assertEqual(md2.target_branch, "http://example.com")

    def test_message(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md3 = self.from_objects(
            tree_a.branch.repository,
            b"rev1",
            500,
            120,
            tree_b.branch.base,
            patch_type=None,
            public_branch=branch_c.base,
            message="Merge message",
        )
        md3.to_lines()
        self.assertIs(None, md3.patch)
        self.assertEqual("Merge message", md3.message)

    def test_generate_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md1 = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            120,
            tree_b.branch.base,
            public_branch=branch_c.base,
        )

        self.assertContainsRe(md1.get_raw_bundle(), b"Bazaar revision bundle")
        self.assertContainsRe(md1.patch, b"\\+content_c")
        self.assertNotContainsRe(md1.patch, b"\\+content_a")
        self.assertContainsRe(md1.patch, b"\\+content_c")
        self.assertNotContainsRe(md1.patch, b"\\+content_a")

    def test_broken_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md1 = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            120,
            tree_b.branch.base,
            public_branch=branch_c.base,
        )
        lines = md1.to_lines()
        lines = [l.replace(b"\n", b"\r\n") for l in lines]
        md2 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(b"rev2a", md2.revision_id)

    def test_signing(self):
        time = 453
        timezone = 7200

        class FakeBranch:
            def get_config_stack(self):
                return self

        md = self.make_merge_directive(
            b"example:",
            b"sha",
            time,
            timezone,
            "http://example.com",
            source_branch="http://example.org",
            patch=b"booga",
            patch_type="diff",
        )
        old_strategy = gpg.GPGStrategy
        gpg.GPGStrategy = gpg.LoopbackGPGStrategy
        try:
            signed = md.to_signed(FakeBranch())
        finally:
            gpg.GPGStrategy = old_strategy
        self.assertContainsRe(signed, b"^-----BEGIN PSEUDO-SIGNED CONTENT")
        self.assertContainsRe(signed, b"example.org")
        self.assertContainsRe(signed, b"booga")

    def test_email(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            476,
            60,
            tree_b.branch.base,
            patch_type=None,
            public_branch=tree_a.branch.base,
        )
        message = md.to_email("pqm@example.com", tree_a.branch)
        self.assertContainsRe(message.as_string(), self.EMAIL1)
        md.message = "Commit of rev2a with special message"
        message = md.to_email("pqm@example.com", tree_a.branch)
        self.assertContainsRe(message.as_string(), self.EMAIL2)

    def test_install_revisions_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type=None,
            public_branch=tree_a.branch.base,
        )
        self.assertFalse(tree_b.branch.repository.has_revision(b"rev2a"))
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b"rev2a", revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b"rev2a"))

    def test_get_merge_request(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type="bundle",
            public_branch=tree_a.branch.base,
        )
        self.assertFalse(tree_b.branch.repository.has_revision(b"rev2a"))
        md.install_revisions(tree_b.branch.repository)
        base, revision, verified = md.get_merge_request(tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual("inapplicable", verified)
        else:
            self.assertEqual(b"rev1", base)
            self.assertEqual("verified", verified)
        self.assertEqual(b"rev2a", revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b"rev2a"))
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type=None,
            public_branch=tree_a.branch.base,
        )
        base, revision, verified = md.get_merge_request(tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual("inapplicable", verified)
        else:
            self.assertEqual(b"rev1", base)
            self.assertEqual("inapplicable", verified)
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type="diff",
            public_branch=tree_a.branch.base,
        )
        base, revision, verified = md.get_merge_request(tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual("inapplicable", verified)
        else:
            self.assertEqual(b"rev1", base)
            self.assertEqual("verified", verified)
        md.patch = b"asdf"
        base, revision, verified = md.get_merge_request(tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual("inapplicable", verified)
        else:
            self.assertEqual(b"rev1", base)
            self.assertEqual("failed", verified)

    def test_install_revisions_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type="bundle",
            public_branch=tree_a.branch.base,
        )
        self.assertFalse(tree_b.branch.repository.has_revision(b"rev2a"))
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b"rev2a", revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b"rev2a"))

    def test_get_target_revision_nofetch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_b.branch.fetch(tree_a.branch)
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            36,
            tree_b.branch.base,
            patch_type=None,
            public_branch=tree_a.branch.base,
        )
        md.source_branch = "/dev/null"
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b"rev2a", revision)

    def test_use_submit_for_missing_dependency(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        self.build_tree_contents([("tree_a/file", b"content_q\ncontent_r\n")])
        tree_a.commit("rev3a", rev_id=b"rev3a")
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev3a",
            500,
            36,
            branch_c.base,
            base_revision_id=b"rev2a",
        )
        md.install_revisions(tree_b.branch.repository)

    def test_handle_target_not_a_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        self.build_tree_contents([("tree_a/file", b"content_q\ncontent_r\n")])
        tree_a.commit("rev3a", rev_id=b"rev3a")
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev3a",
            500,
            36,
            branch_c.base,
            base_revision_id=b"rev2a",
        )
        md.target_branch = self.get_url("not-a-branch")
        self.assertRaises(
            errors.TargetNotBranch, md.install_revisions, tree_b.branch.repository
        )


class TestMergeDirective1Branch(tests.TestCaseWithTransport, TestMergeDirectiveBranch):
    """Test merge directive format 1 with a branch."""

    EMAIL1 = EMAIL1

    EMAIL2 = EMAIL2

    def from_objects(
        self,
        repository,
        revision_id,
        time,
        timezone,
        target_branch,
        patch_type="bundle",
        local_target_branch=None,
        public_branch=None,
        message=None,
        base_revision_id=None,
    ):
        if base_revision_id is not None:
            raise tests.TestNotApplicable(
                "This format does not support explicit bases."
            )
        with repository.lock_write():
            return merge_directive.MergeDirective.from_objects(
                repository,
                revision_id,
                time,
                timezone,
                target_branch,
                patch_type,
                local_target_branch,
                public_branch,
                message,
            )

    def make_merge_directive(
        self,
        revision_id,
        testament_sha1,
        time,
        timezone,
        target_branch,
        patch=None,
        patch_type=None,
        source_branch=None,
        message=None,
    ):
        return merge_directive.MergeDirective(
            revision_id,
            testament_sha1,
            time,
            timezone,
            target_branch,
            patch,
            patch_type,
            source_branch,
            message,
        )


class TestMergeDirective2Branch(tests.TestCaseWithTransport, TestMergeDirectiveBranch):
    """Test merge directive format 2 with a branch."""

    EMAIL1 = EMAIL1_2

    EMAIL2 = EMAIL2_2

    def from_objects(
        self,
        repository,
        revision_id,
        time,
        timezone,
        target_branch,
        patch_type="bundle",
        local_target_branch=None,
        public_branch=None,
        message=None,
        base_revision_id=None,
    ):
        include_patch = patch_type in ("bundle", "diff")
        include_bundle = patch_type == "bundle"
        self.assertTrue(patch_type in ("bundle", "diff", None))
        return merge_directive.MergeDirective2.from_objects(
            repository,
            revision_id,
            time,
            timezone,
            target_branch,
            include_patch,
            include_bundle,
            local_target_branch,
            public_branch,
            message,
            base_revision_id,
        )

    def make_merge_directive(
        self,
        revision_id,
        testament_sha1,
        time,
        timezone,
        target_branch,
        patch=None,
        patch_type=None,
        source_branch=None,
        message=None,
        base_revision_id=b"null:",
    ):
        if patch_type == "bundle":
            bundle = patch
            patch = None
        else:
            bundle = None
        return merge_directive.MergeDirective2(
            revision_id,
            testament_sha1,
            time,
            timezone,
            target_branch,
            patch,
            source_branch,
            message,
            bundle,
            base_revision_id,
        )

    def test_base_revision(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            60,
            tree_b.branch.base,
            patch_type="bundle",
            public_branch=tree_a.branch.base,
            base_revision_id=None,
        )
        self.assertEqual(b"rev1", md.base_revision_id)
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            60,
            tree_b.branch.base,
            patch_type="bundle",
            public_branch=tree_a.branch.base,
            base_revision_id=b"null:",
        )
        self.assertEqual(b"null:", md.base_revision_id)
        lines = md.to_lines()
        md2 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(md2.base_revision_id, md.base_revision_id)

    def test_patch_verification(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(
            tree_a.branch.repository,
            b"rev2a",
            500,
            60,
            tree_b.branch.base,
            patch_type="bundle",
            public_branch=tree_a.branch.base,
        )
        lines = md.to_lines()
        md2 = merge_directive.MergeDirective.from_lines(lines)
        md2._verify_patch(tree_a.branch.repository)
        # Strip trailing whitespace
        md2.patch = md2.patch.replace(b" \n", b"\n")
        md2._verify_patch(tree_a.branch.repository)
        # Convert to Mac line-endings
        md2.patch = re.sub(b"(\r\n|\r|\n)", b"\r", md2.patch)
        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
        # Convert to DOS line-endings
        md2.patch = re.sub(b"(\r\n|\r|\n)", b"\r\n", md2.patch)
        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
        md2.patch = md2.patch.replace(b"content_c", b"content_d")
        self.assertFalse(md2._verify_patch(tree_a.branch.repository))


class TestParseOldMergeDirective2(tests.TestCase):
    def test_parse_old_merge_directive(self):
        md = merge_directive.MergeDirective.from_lines(INPUT1_2_OLD)
        self.assertEqual(b"example:", md.revision_id)
        self.assertEqual(b"sha", md.testament_sha1)
        self.assertEqual("http://example.com", md.target_branch)
        self.assertEqual("http://example.org", md.source_branch)
        self.assertEqual(453, md.time)
        self.assertEqual(120, md.timezone)
        self.assertEqual(b"booga", md.patch)
        self.assertEqual("diff", md.patch_type)
        self.assertEqual("Hi mom!", md.message)


class TestHook:
    """Hook callback for test purposes."""

    def __init__(self, result=None):
        self.calls = []
        self.result = result

    def __call__(self, params):
        self.calls.append(params)
        return self.result


class HookMailClient(mail_client.MailClient):
    """Mail client for testing hooks."""

    def __init__(self, config):
        self.body = None
        self.config = config

    def compose(
        self,
        prompt,
        to,
        subject,
        attachment,
        mime_subtype,
        extension,
        basename=None,
        body=None,
    ):
        self.body = body


class TestBodyHook(tests.TestCaseWithTransport):
    def compose_with_hooks(self, test_hooks, supports_body=True):
        client = HookMailClient({})
        client.supports_body = supports_body
        for test_hook in test_hooks:
            merge_directive.MergeDirective.hooks.install_named_hook(
                "merge_request_body", test_hook, "test"
            )
        tree = self.make_branch_and_tree("foo")
        tree.commit("foo")
        directive = merge_directive.MergeDirective2(
            tree.branch.last_revision(),
            b"sha",
            0,
            0,
            b"sha",
            source_branch=tree.branch.base,
            base_revision_id=tree.branch.last_revision(),
            message="This code rox",
        )
        directive.compose_merge_request(
            client, "jrandom@example.com", None, tree.branch
        )
        return client, directive

    def test_no_supports_body(self):
        test_hook = TestHook("foo")
        old_warn = trace.warning
        warnings = []

        def warn(*args):
            warnings.append(args)

        trace.warning = warn
        try:
            client, directive = self.compose_with_hooks(
                [test_hook], supports_body=False
            )
        finally:
            trace.warning = old_warn
        self.assertEqual(0, len(test_hook.calls))
        self.assertEqual(
            (
                "Cannot run merge_request_body hooks because mail"
                " client %s does not support message bodies.",
                "HookMailClient",
            ),
            warnings[0],
        )

    def test_body_hook(self):
        test_hook = TestHook("foo")
        client, directive = self.compose_with_hooks([test_hook])
        self.assertEqual(1, len(test_hook.calls))
        self.assertEqual("foo", client.body)
        params = test_hook.calls[0]
        self.assertIsInstance(params, merge_directive.MergeRequestBodyParams)
        self.assertIs(None, params.body)
        self.assertIs(None, params.orig_body)
        self.assertEqual("jrandom@example.com", params.to)
        self.assertEqual("[MERGE] This code rox", params.subject)
        self.assertEqual(directive, params.directive)
        self.assertEqual("foo-1", params.basename)

    def test_body_hook_chaining(self):
        test_hook1 = TestHook("foo")
        test_hook2 = TestHook("bar")
        client = self.compose_with_hooks([test_hook1, test_hook2])[0]
        self.assertEqual(None, test_hook1.calls[0].body)
        self.assertEqual(None, test_hook1.calls[0].orig_body)
        self.assertEqual("foo", test_hook2.calls[0].body)
        self.assertEqual(None, test_hook2.calls[0].orig_body)
        self.assertEqual("bar", client.body)



#! /usr/bin/env python

# Originally by Dave Goodger, from the docutils, distribution.
#
# Modified for Bazaar to accommodate options containing dots
#
# This file is in the public domain.

"""A minimal front end to the Docutils Publisher, producing HTML."""

try:
    import locale

    locale.setlocale(locale.LC_ALL, "")
except BaseException:
    pass

from docutils.core import default_description, publish_cmdline

if True:  # this is still required in the distutils trunk as-at June 2008.
    from docutils.parsers.rst.states import Body

    # we have some option names that contain dot; which is not allowed by
    # python-docutils 0.4-4 -- so monkeypatch in a better pattern
    #
    # This is a bit gross to patch because all this is built up at load time.
    Body.pats["optname"] = r"[a-zA-Z0-9][a-zA-Z0-9._-]*"
    Body.pats["longopt"] = r"(--|/){optname}([ =]{optarg})?".format(**Body.pats)
    Body.pats["option"] = r"({shortopt}|{longopt})".format(**Body.pats)
    Body.patterns["option_marker"] = r"{option}(, {option})*(  +| ?$)".format(
        **Body.pats
    )


description = (
    "Generates (X)HTML documents from standalone reStructuredText "
    "sources.  " + default_description
)


# workaround for bug with <xxx id="tags" name="tags"> in IE
from docutils.writers import html4css1


class IESafeHtmlTranslator(html4css1.HTMLTranslator):
    def starttag(self, node, tagname, suffix="\n", empty=0, **attributes):
        x = html4css1.HTMLTranslator.starttag(
            self, node, tagname, suffix, empty, **attributes
        )
        y = x.replace('id="tags"', 'id="tags_"')
        y = y.replace('name="tags"', 'name="tags_"')
        y = y.replace('href="#tags"', 'href="#tags_"')
        return y


mywriter = html4css1.Writer()
mywriter.translator_class = IESafeHtmlTranslator


publish_cmdline(writer=mywriter, description=description)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.add'."""

from breezy import errors, tests
from breezy.bzr import inventory
from breezy.tests.matchers import HasLayout, HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestAdd(TestCaseWithWorkingTree):
    def assertTreeLayout(self, expected, tree):
        """Check that the tree has the correct layout."""
        self.assertThat(tree, HasLayout(expected))

    def assertPathRelations(self, previous_tree, tree, relations):
        self.assertThat(tree, HasPathRelations(previous_tree, relations))

    def test_add_one(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["one"])
        tree.add("one")

        self.assertTreeLayout(["", "one"], tree)

    def test_add_existing_id(self):
        """Adding an entry with a pre-existing id raises DuplicateFileId."""
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["a", "b"])
        tree.add(["a"])
        self.assertRaises(
            inventory.DuplicateFileId, tree.add, ["b"], ids=[tree.path2id("a")]
        )
        # And the entry should not have been added.
        self.assertTreeLayout(["", "a"], tree)

    def test_add_old_id(self):
        """We can add an old id, as long as it doesn't exist now."""
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["a", "b"])
        tree.add(["a"])
        file_id = tree.path2id("a")
        tree.commit("first")
        # And the entry should not have been added.
        tree.unversion(["a"])
        tree.add(["b"], ids=[file_id])
        self.assertPathRelations(tree.basis_tree(), tree, [("", ""), ("b", "a")])

    def test_add_one_list(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["one"])
        tree.add(["one"])

        self.assertTreeLayout(["", "one"], tree)

    def test_add_one_new_id(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["one"])
        tree.add(["one"])

        self.assertTreeLayout(["", "one"], tree)

    def test_add_unicode(self):
        tree = self.make_branch_and_tree(".")
        try:
            self.build_tree(["f\xf6"])
        except UnicodeError:
            raise tests.TestSkipped("Filesystem does not support filename.")
        tree.add(["f\xf6"])

        self.assertTreeLayout(["", "f\xf6"], tree)

    def test_add_subdir_with_ids(self):
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.add(["dir"], ids=[b"dir-id"])
        tree.add(["dir/subdir"], ids=[b"subdir-id"])
        tree.add(["dir/subdir/foo"], ids=[b"foo-id"])
        root_id = tree.path2id("")

        self.assertTreeLayout(
            [
                ("", root_id),
                ("dir/", b"dir-id"),
                ("dir/subdir/", b"subdir-id"),
                ("dir/subdir/foo", b"foo-id"),
            ],
            tree,
        )

    def test_add_subdir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.add(["dir"])
        tree.add(["dir/subdir"])
        tree.add(["dir/subdir/foo"])

        self.assertTreeLayout(["", "dir/", "dir/subdir/", "dir/subdir/foo"], tree)

    def test_add_multiple(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b", "dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.add(["a", "b", "dir", "dir/subdir", "dir/subdir/foo"])

        self.assertTreeLayout(
            ["", "a", "b", "dir/", "dir/subdir/", "dir/subdir/foo"], tree
        )

    def test_add_multiple_with_file_ids(self):
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["a", "b", "dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.add(
            ["a", "b", "dir", "dir/subdir", "dir/subdir/foo"],
            ids=[b"a-id", b"b-id", b"dir-id", b"subdir-id", b"foo-id"],
        )

        self.assertTreeLayout(
            [
                ("", tree.path2id("")),
                ("a", b"a-id"),
                ("b", b"b-id"),
                ("dir/", b"dir-id"),
                ("dir/subdir/", b"subdir-id"),
                ("dir/subdir/foo", b"foo-id"),
            ],
            tree,
        )

    def test_add_invalid(self):
        tree = self.make_branch_and_tree(".")
        if not tree._format.supports_versioned_directories:
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )
        self.build_tree(["dir/", "dir/subdir/", "dir/subdir/foo"])

        self.assertRaises(errors.NotVersionedError, tree.add, ["dir/subdir"])
        self.assertTreeLayout([""], tree)

    def test_add_after_remove(self):
        tree = self.make_branch_and_tree(".")
        if not tree._format.supports_versioned_directories:
            raise tests.TestNotApplicable(
                "format does not support versioned directories"
            )
        self.build_tree(["dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.add(["dir"])
        tree.commit("dir")
        tree.unversion(["dir"])
        self.assertRaises(errors.NotVersionedError, tree.add, ["dir/subdir/foo"])

    def test_add_root(self):
        # adding the root should be a no-op, or at least not
        # do anything whacky.
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            tree.add("")
            self.assertEqual([""], list(tree.all_versioned_paths()))
            # the root should have been changed to be a new unique root.
            if tree._format.supports_setting_file_ids:
                self.assertNotEqual(inventory.ROOT_ID, tree.path2id(""))

    def test_add_previously_added(self):
        # adding a path that was previously added should work
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo"])
        tree.add(["foo"])
        tree.unversion(["foo"])
        tree.add(["foo"])
        self.assertTrue(tree.has_filename("foo"))

    def test_add_previously_added_with_file_id(self):
        # adding a path that was previously added should work
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["foo"])
        tree.add(["foo"], ids=[b"foo-id"])
        tree.unversion(["foo"])
        tree.add(["foo"], ids=[b"foo-id"])
        self.assertEqual(b"foo-id", tree.path2id("foo"))

    def test_add_present_in_basis(self):
        # adding a path that was present in the basis should work.
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo"])
        tree.add(["foo"])
        tree.commit("add foo")
        tree.unversion(["foo"])
        tree.add(["foo"])
        self.assertTrue(tree.has_filename("foo"))

    def test_add_present_in_basis_with_file_ids(self):
        # adding a path that was present in the basis should work.
        tree = self.make_branch_and_tree(".")
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(["foo"])
        tree.add(["foo"], ids=[b"foo-id"])
        tree.commit("add foo")
        tree.unversion(["foo"])
        tree.add(["foo"], ids=[b"foo-id"])
        self.assertEqual(b"foo-id", tree.path2id("foo"))



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for InterRepository implementastions."""

import sys

import breezy
import breezy.errors as errors
import breezy.gpg
from breezy.bzr.inventory import Inventory
from breezy.repository import WriteGroup
from breezy.revision import NULL_REVISION
from breezy.tests import TestNotApplicable, TestSkipped
from breezy.tests.matchers import MatchesAncestry
from breezy.tests.per_interrepository import TestCaseWithInterRepository
from breezy.workingtree import WorkingTree


def check_repo_format_for_funky_id_on_win32(repo):
    if not repo._format.supports_funky_characters and sys.platform == "win32":
        raise TestSkipped(
            "funky chars not allowed on this platform in repository {}".format(
                repo.__class__.__name__
            )
        )


class TestInterRepository(TestCaseWithInterRepository):
    def test_interrepository_get_returns_correct_optimiser(self):
        # we assume the optimising code paths are triggered
        # by the type of the repo not the transport - at this point.
        # we may need to update this test if this changes.
        #
        # XXX: This code tests that we get an InterRepository when we try to
        # convert between the two repositories that it wants to be tested with
        # -- but that's not necessarily correct.  So for now this is disabled.
        # mbp 20070206
        ## source_repo = self.make_repository("source")
        ## target_repo = self.make_to_repository("target")
        ## interrepo = repository.InterRepository.get(source_repo, target_repo)
        ## self.assertEqual(self.interrepo_class, interrepo.__class__)
        pass


class TestCaseWithComplexRepository(TestCaseWithInterRepository):
    def setUp(self):
        super().setUp()
        tree_a = self.make_branch_and_tree("a")
        self.controldir = tree_a.branch.controldir
        # add a corrupt inventory 'orphan'
        with (
            tree_a.branch.repository.lock_write(),
            WriteGroup(tree_a.branch.repository),
        ):
            if tree_a.branch.repository._format.supports_ghosts:
                inv_file = tree_a.branch.repository.inventories
                inv_file.add_lines((b"orphan",), [], [])
        # add a real revision 'rev1'
        self.rev1 = tree_a.commit("rev1", allow_pointless=True)
        # add a real revision 'rev2' based on rev1
        self.rev2 = tree_a.commit("rev2", allow_pointless=True)

    def test_search_missing_revision_ids(self):
        # revision ids in repository A but not B are returned, fake ones
        # are stripped. (fake meaning no revision object, but an inventory
        # as some formats keyed off inventory data in the past.)
        # make a repository to compare against that claims to have rev1
        repo_b = self.make_to_repository("rev1_only")
        repo_a = self.controldir.open_repository()
        try:
            repo_b.fetch(repo_a, self.rev1)
        except errors.NoRoundtrippingSupport:
            raise TestNotApplicable("roundtripping not supported")
        # check the test will be valid
        self.assertFalse(repo_b.has_revision(self.rev2))
        result = repo_b.search_missing_revision_ids(repo_a)
        self.assertEqual({self.rev2}, result.get_keys())
        self.assertEqual(("search", {self.rev2}, {self.rev1}, 1), result.get_recipe())

    def test_absent_requested_raises(self):
        # Asking for missing revisions with a tip that is itself absent in the
        # source raises NoSuchRevision.
        repo_b = self.make_to_repository("target")
        repo_a = self.controldir.open_repository()
        # No pizza revisions anywhere
        self.assertFalse(repo_a.has_revision(b"pizza"))
        self.assertFalse(repo_b.has_revision(b"pizza"))
        # Asking specifically for an absent revision errors.
        self.assertRaises(
            errors.NoSuchRevision,
            repo_b.search_missing_revision_ids,
            repo_a,
            revision_ids=[b"pizza"],
            find_ghosts=True,
        )
        self.assertRaises(
            errors.NoSuchRevision,
            repo_b.search_missing_revision_ids,
            repo_a,
            revision_ids=[b"pizza"],
            find_ghosts=False,
        )

    def test_search_missing_rev_limited(self):
        # revision ids in repository A that are not referenced by the
        # requested revision are not returned.
        # make a repository to compare against that is empty
        repo_b = self.make_to_repository("empty")
        repo_a = self.controldir.open_repository()
        result = repo_b.search_missing_revision_ids(repo_a, revision_ids=[self.rev1])
        self.assertEqual({self.rev1}, result.get_keys())
        self.assertEqual(
            ("search", {self.rev1}, {NULL_REVISION}, 1), result.get_recipe()
        )

    def test_search_missing_revision_ids_limit(self):
        # The limit= argument makes fetch() limit
        # the results to the first X topo-sorted revisions.
        repo_b = self.make_to_repository("rev1_only")
        repo_a = self.controldir.open_repository()
        # check the test will be valid
        self.assertFalse(repo_b.has_revision(self.rev2))
        try:
            result = repo_b.search_missing_revision_ids(repo_a, limit=1)
        except errors.FetchLimitUnsupported:
            raise TestNotApplicable("interrepo does not support limited fetches")
        self.assertEqual(("search", {self.rev1}, {b"null:"}, 1), result.get_recipe())

    def test_fetch_fetches_signatures_too(self):
        if not self.repository_format.supports_revision_signatures:
            raise TestNotApplicable("from repository does not support signatures")
        if not self.repository_format_to.supports_revision_signatures:
            raise TestNotApplicable("to repository does not support signatures")
        # and sign 'rev2'
        tree_a = WorkingTree.open("a")
        with (
            tree_a.branch.repository.lock_write(),
            WriteGroup(tree_a.branch.repository),
        ):
            tree_a.branch.repository.sign_revision(
                self.rev2, breezy.gpg.LoopbackGPGStrategy(None)
            )

        from_repo = self.controldir.open_repository()
        from_signature = from_repo.get_signature_text(self.rev2)
        to_repo = self.make_to_repository("target")
        try:
            to_repo.fetch(from_repo)
        except errors.NoRoundtrippingSupport:
            raise TestNotApplicable("interrepo does not support roundtripping")
        to_signature = to_repo.get_signature_text(self.rev2)
        self.assertEqual(from_signature, to_signature)


class TestCaseWithGhosts(TestCaseWithInterRepository):
    def test_fetch_all_fixes_up_ghost(self):
        # we want two repositories at this point:
        # one with a revision that is a ghost in the other
        # repository.
        # 'ghost' is present in has_ghost, 'ghost' is absent in 'missing_ghost'.
        # 'references' is present in both repositories, and 'tip' is present
        # just in has_ghost.
        # has_ghost       missing_ghost
        # ------------------------------
        # 'ghost'             -
        # 'references'    'references'
        # 'tip'               -
        # In this test we fetch 'tip' which should not fetch 'ghost'
        has_ghost = self.make_repository("has_ghost")
        missing_ghost = self.make_repository("missing_ghost")
        if [repo._format.supports_ghosts for repo in (has_ghost, missing_ghost)] != [
            True,
            True,
        ]:
            raise TestNotApplicable("Need ghost support.")

        def add_commit(repo, revision_id, parent_ids):
            repo.lock_write()
            repo.start_write_group()
            inv = Inventory(revision_id=revision_id)
            inv.root.revision = revision_id
            root_id = inv.root.file_id
            sha1 = repo.add_inventory(revision_id, inv, parent_ids)
            repo.texts.add_lines((root_id, revision_id), [], [])
            rev = breezy.revision.Revision(
                timestamp=0,
                timezone=None,
                committer="Foo Bar <foo@example.com>",
                message="Message",
                inventory_sha1=sha1,
                revision_id=revision_id,
            )
            rev.parent_ids = parent_ids
            repo.add_revision(revision_id, rev)
            repo.commit_write_group()
            repo.unlock()

        add_commit(has_ghost, b"ghost", [])
        add_commit(has_ghost, b"references", [b"ghost"])
        add_commit(missing_ghost, b"references", [b"ghost"])
        add_commit(has_ghost, b"tip", [b"references"])
        missing_ghost.fetch(has_ghost, b"tip", find_ghosts=True)
        # missing ghost now has tip and ghost.
        missing_ghost.get_revision(b"tip")
        missing_ghost.get_inventory(b"tip")
        missing_ghost.get_revision(b"ghost")
        missing_ghost.get_inventory(b"ghost")
        # rev must not be corrupt now
        self.assertThat(
            [b"ghost", b"references", b"tip"], MatchesAncestry(missing_ghost, b"tip")
        )



# Copyright (C) 2010 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for managing external merge tools such as kdiff3."""

import os
import shutil
import subprocess
import sys
import tempfile

from .lazy_import import lazy_import

lazy_import(
    globals(),
    """
from breezy import (
    cmdline,
    osutils,
    trace,
)
""",
)


known_merge_tools = {
    "bcompare": "bcompare {this} {other} {base} {result}",
    "kdiff3": "kdiff3 {base} {this} {other} -o {result}",
    "xdiff": "xxdiff -m -O -M {result} {this} {base} {other}",
    "meld": "meld {base} {this_temp} {other}",
    "opendiff": "opendiff {this} {other} -ancestor {base} -merge {result}",
    "winmergeu": "winmergeu {result}",
}


def check_availability(command_line):
    cmd_list = cmdline.split(command_line)
    exe = cmd_list[0]
    if sys.platform == "win32":
        exe = _get_executable_path(exe)
        if exe is None:
            return False
        base, ext = os.path.splitext(exe)
        path_ext = [s.lower() for s in os.getenv("PATHEXT", "").split(os.pathsep)]
        return os.path.exists(exe) and ext in path_ext
    else:
        return (
            os.access(exe, os.X_OK) or osutils.find_executable_on_path(exe) is not None
        )


def invoke(command_line, filename, invoker=None):
    """Invokes the given merge tool command line, substituting the given
    filename according to the embedded substitution markers. Optionally, it
    will use the given invoker function instead of the default
    subprocess_invoker.
    """
    if invoker is None:
        invoker = subprocess_invoker
    cmd_list = cmdline.split(command_line)
    exe = _get_executable_path(cmd_list[0])
    if exe is not None:
        cmd_list[0] = exe
    args, tmp_file = _subst_filename(cmd_list, filename)

    def cleanup(retcode):
        if tmp_file is not None:
            if retcode == 0:  # on success, replace file with temp file
                shutil.move(tmp_file, filename)
            else:  # otherwise, delete temp file
                os.remove(tmp_file)

    return invoker(args[0], args[1:], cleanup)


def _get_executable_path(exe):
    if os.path.isabs(exe):
        return exe
    return osutils.find_executable_on_path(exe)


def _subst_filename(args, filename):
    subst_names = {
        "base": filename + ".BASE",
        "this": filename + ".THIS",
        "other": filename + ".OTHER",
        "result": filename,
    }
    tmp_file = None
    subst_args = []
    for arg in args:
        if "{this_temp}" in arg and "this_temp" not in subst_names:
            fh, tmp_file = tempfile.mkstemp(
                "_bzr_mergetools_{}.THIS".format(os.path.basename(filename))
            )
            trace.mutter("fh=%r, tmp_file=%r", fh, tmp_file)
            os.close(fh)
            shutil.copy(filename + ".THIS", tmp_file)
            subst_names["this_temp"] = tmp_file
        arg = _format_arg(arg, subst_names)
        subst_args.append(arg)
    return subst_args, tmp_file


# This would be better implemented using format() from python 2.6
def _format_arg(arg, subst_names):
    arg = arg.replace("{base}", subst_names["base"])
    arg = arg.replace("{this}", subst_names["this"])
    arg = arg.replace("{other}", subst_names["other"])
    arg = arg.replace("{result}", subst_names["result"])
    if "this_temp" in subst_names:
        arg = arg.replace("{this_temp}", subst_names["this_temp"])
    return arg


def subprocess_invoker(executable, args, cleanup):
    retcode = subprocess.call([executable] + args)
    cleanup(retcode)
    return retcode



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import tarfile
import zipfile

from breezy import osutils
from breezy.errors import UnsupportedOperation
from breezy.export import export
from breezy.tests import TestNotApplicable, features
from breezy.tests.per_tree import TestCaseWithTree


class ExportTest:
    def prepare_export(self):
        work_a = self.make_branch_and_tree("wta")
        self.build_tree_contents([("wta/file", b"a\nb\nc\nd\n"), ("wta/dir", b"")])
        work_a.add("file")
        work_a.add("dir")
        work_a.commit("add file")
        tree_a = self.workingtree_to_test_tree(work_a)
        export(tree_a, "output", self.exporter)

    def prepare_symlink_export(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        work_a = self.make_branch_and_tree("wta")
        os.symlink("target", "wta/link")
        work_a.add("link")
        work_a.commit("add link")
        tree_a = self.workingtree_to_test_tree(work_a)
        export(tree_a, "output", self.exporter)

    def test_export(self):
        self.prepare_export()
        names = self.get_export_names()
        self.assertIn("output/file", names)
        self.assertIn("output/dir", names)

    def test_export_symlink(self):
        self.prepare_symlink_export()
        names = self.get_export_names()
        self.assertIn("output/link", names)

    def prepare_nested_export(self, recurse_nested):
        tree = self.make_branch_and_tree("dir")
        self.build_tree(["dir/a"])
        tree.add("a")
        tree.commit("1")

        subtree = self.make_branch_and_tree("dir/subdir")

        self.build_tree(["dir/subdir/b"])
        subtree.add("b")
        subtree.commit("1a")

        try:
            tree.add_reference(subtree)
        except UnsupportedOperation:
            raise TestNotApplicable("format does not supported nested trees")
        tree.commit("2")
        export(tree, "output", self.exporter, recurse_nested=recurse_nested)

    def test_export_nested_recurse(self):
        self.prepare_nested_export(True)
        names = self.get_export_names()
        self.assertIn("output/subdir/b", names)

    def test_export_nested_nonrecurse(self):
        self.prepare_nested_export(False)
        names = self.get_export_names()
        self.assertNotIn("output/subdir/b", names)


class TestTar(ExportTest, TestCaseWithTree):
    exporter = "tar"

    def get_export_names(self):
        tf = tarfile.open("output")
        try:
            return tf.getnames()
        finally:
            tf.close()


class TestZip(ExportTest, TestCaseWithTree):
    exporter = "zip"

    def get_export_names(self):
        zf = zipfile.ZipFile("output")
        try:
            return zf.namelist()
        finally:
            zf.close()

    def test_export_symlink(self):
        self.prepare_symlink_export()
        names = self.get_export_names()
        self.assertIn("output/link.lnk", names)


class TestDir(ExportTest, TestCaseWithTree):
    exporter = "dir"

    def get_export_names(self):
        ret = []
        for dirpath, dirnames, filenames in os.walk("output"):
            for dirname in dirnames:
                ret.append(osutils.pathjoin(dirpath, dirname))
            for filename in filenames:
                ret.append(osutils.pathjoin(dirpath, filename))
        return ret



# Copyright (C) 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Direct tests of the revision_store classes."""

from .... import errors, osutils, tests
from ....bzr import inventory
from .. import revision_store
from . import FastimportFeature


class Test_TreeShim(tests.TestCase):
    _test_needs_features = [FastimportFeature]

    def invAddEntry(self, inv, path, file_id=None):
        if path.endswith("/"):
            path = path[:-1]
            kind = "directory"
        else:
            kind = "file"
        parent_path, basename = osutils.split(path)
        parent_id = inv.path2id(parent_path)
        inv.add(inventory.make_entry(kind, basename, parent_id, file_id))

    def make_trivial_basis_inv(self):
        basis_inv = inventory.Inventory(b"TREE_ROOT")
        self.invAddEntry(basis_inv, "foo", b"foo-id")
        self.invAddEntry(basis_inv, "bar/", b"bar-id")
        self.invAddEntry(basis_inv, "bar/baz", b"baz-id")
        return basis_inv

    def test_id2path_no_delta(self):
        basis_inv = self.make_trivial_basis_inv()
        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=[], content_provider=None
        )
        self.assertEqual("", shim.id2path(b"TREE_ROOT"))
        self.assertEqual("foo", shim.id2path(b"foo-id"))
        self.assertEqual("bar", shim.id2path(b"bar-id"))
        self.assertEqual("bar/baz", shim.id2path(b"baz-id"))
        self.assertRaises(errors.NoSuchId, shim.id2path, b"qux-id")

    def test_id2path_with_delta(self):
        basis_inv = self.make_trivial_basis_inv()
        foo_entry = inventory.make_entry("file", "foo2", b"TREE_ROOT", b"foo-id")
        inv_delta = [
            ("foo", "foo2", b"foo-id", foo_entry),
            ("bar/baz", None, b"baz-id", None),
        ]

        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=inv_delta, content_provider=None
        )
        self.assertEqual("", shim.id2path(b"TREE_ROOT"))
        self.assertEqual("foo2", shim.id2path(b"foo-id"))
        self.assertEqual("bar", shim.id2path(b"bar-id"))
        self.assertRaises(errors.NoSuchId, shim.id2path, b"baz-id")

    def test_path2id(self):
        basis_inv = self.make_trivial_basis_inv()
        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=[], content_provider=None
        )
        self.assertEqual(b"TREE_ROOT", shim.path2id(""))
        self.assertEqual(b"bar-id", shim.path2id("bar"))

    def test_get_file_with_stat_content_in_stream(self):
        basis_inv = self.make_trivial_basis_inv()

        def content_provider(file_id):
            return b"content of\n" + file_id + b"\n"

        shim = revision_store._TreeShim(
            repo=None,
            basis_inv=basis_inv,
            inv_delta=[],
            content_provider=content_provider,
        )
        f_obj, stat_val = shim.get_file_with_stat("bar/baz")
        self.assertIs(None, stat_val)
        self.assertEqualDiff(b"content of\nbaz-id\n", f_obj.read())

    # TODO: Test when the content isn't in the stream, and we fall back to the
    #       repository that was passed in

    def test_get_symlink_target(self):
        basis_inv = self.make_trivial_basis_inv()
        ie = inventory.make_entry("symlink", "link", b"TREE_ROOT", b"link-id")
        ie.symlink_target = "link-target"
        basis_inv.add(ie)
        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=[], content_provider=None
        )
        self.assertEqual("link-target", shim.get_symlink_target("link"))

    def test_get_symlink_target_from_delta(self):
        basis_inv = self.make_trivial_basis_inv()
        ie = inventory.make_entry("symlink", "link", b"TREE_ROOT", b"link-id")
        ie.symlink_target = "link-target"
        inv_delta = [(None, "link", b"link-id", ie)]
        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=inv_delta, content_provider=None
        )
        self.assertEqual("link-target", shim.get_symlink_target("link"))

    def test__delta_to_iter_changes(self):
        basis_inv = self.make_trivial_basis_inv()
        foo_entry = inventory.make_entry("file", "foo2", b"bar-id", b"foo-id")
        link_entry = inventory.make_entry("symlink", "link", b"TREE_ROOT", b"link-id")
        link_entry.symlink_target = "link-target"
        inv_delta = [
            ("foo", "bar/foo2", b"foo-id", foo_entry),
            ("bar/baz", None, b"baz-id", None),
            (None, "link", b"link-id", link_entry),
        ]
        shim = revision_store._TreeShim(
            repo=None, basis_inv=basis_inv, inv_delta=inv_delta, content_provider=None
        )
        changes = list(shim._delta_to_iter_changes())
        expected = [
            (
                b"foo-id",
                ("foo", "bar/foo2"),
                False,
                (True, True),
                (b"TREE_ROOT", b"bar-id"),
                ("foo", "foo2"),
                ("file", "file"),
                (False, False),
                False,
            ),
            (
                b"baz-id",
                ("bar/baz", None),
                True,
                (True, False),
                (b"bar-id", None),
                ("baz", None),
                ("file", None),
                (False, None),
                False,
            ),
            (
                b"link-id",
                (None, "link"),
                True,
                (False, True),
                (None, b"TREE_ROOT"),
                (None, "link"),
                (None, "symlink"),
                (None, False),
                False,
            ),
        ]
        # from pprint import pformat
        # self.assertEqualDiff(pformat(expected), pformat(changes))
        self.assertEqual(expected, changes)



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black box tests for the upgrade ui."""

import os
import stat

from breezy import bzr, controldir, lockdir, ui, urlutils
from breezy.bzr import bzrdir
from breezy.bzr.knitpack_repo import RepositoryFormatKnitPack1
from breezy.tests import TestCaseWithTransport, features
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer


class OldBzrDir(bzrdir.BzrDirMeta1):
    """An test brz dir implementation."""

    def needs_format_conversion(self, format):
        return not isinstance(format, self.__class__)


class ConvertOldTestToMeta(controldir.Converter):
    """A trivial converter, used for testing."""

    def convert(self, to_convert, pb):
        ui.ui_factory.note("starting upgrade from old test format to 2a")
        to_convert.control_transport.put_bytes(
            "branch-format",
            bzrdir.BzrDirMetaFormat1().get_format_string(),
            mode=to_convert._get_file_mode(),
        )
        return controldir.ControlDir.open(to_convert.user_url)


class OldBzrDirFormat(bzrdir.BzrDirMetaFormat1):
    _lock_class = lockdir.LockDir

    def get_converter(self, format=None):
        return ConvertOldTestToMeta()

    @classmethod
    def get_format_string(cls):
        return b"Ancient Test Format"

    def _open(self, transport):
        return OldBzrDir(transport, self)


class TestWithUpgradableBranches(TestCaseWithTransport):
    def setUp(self):
        super().setUp()

    def make_current_format_branch_and_checkout(self):
        current_tree = self.make_branch_and_tree(
            "current_format_branch", format="default"
        )
        current_tree.branch.create_checkout(
            self.get_url("current_format_checkout"), lightweight=True
        )

    def test_readonly_url_error(self):
        self.make_branch_and_tree("old_format_branch", format="knit")
        (out, err) = self.run_bzr(
            ["upgrade", self.get_readonly_url("old_format_branch")], retcode=3
        )
        err_msg = "Upgrade URL cannot work with readonly URLs."
        self.assertEqualDiff(
            "conversion error: {}\nbrz: ERROR: {}\n".format(err_msg, err_msg), err
        )

    def test_upgrade_up_to_date(self):
        self.make_current_format_branch_and_checkout()
        # when up to date we should get a message to that effect
        burl = self.get_transport("current_format_branch").local_abspath(".")
        (out, err) = self.run_bzr("upgrade current_format_branch", retcode=0)
        self.assertEqual(
            "Upgrading branch {}/ ...\n"
            "The branch format {} is already at the most recent format.\n".format(
                burl, "Meta directory format 1"
            ),
            out,
        )

    def test_upgrade_up_to_date_checkout_warns_branch_left_alone(self):
        self.make_current_format_branch_and_checkout()
        # when upgrading a checkout, the branch location and a suggestion
        # to upgrade it should be emitted even if the checkout is up to
        # date
        burl = self.get_transport("current_format_branch").local_abspath(".")
        curl = self.get_transport("current_format_checkout").local_abspath(".")
        (out, err) = self.run_bzr("upgrade current_format_checkout", retcode=0)
        self.assertEqual(
            "Upgrading branch {}/ ...\nThis is a checkout."
            " The branch ({}/) needs to be upgraded separately.\n"
            "The branch format {} is already at the most recent format.\n".format(
                curl, burl, "Meta directory format 1"
            ),
            out,
        )

    def test_upgrade_checkout(self):
        # upgrading a checkout should work
        pass

    def test_upgrade_repository_scans_branches(self):
        # we should get individual upgrade notes for each branch even the
        # anonymous branch
        pass

    def test_upgrade_branch_in_repo(self):
        # upgrading a branch in a repo should warn about not upgrading the repo
        pass

    def test_upgrade_control_dir(self):
        old_format = OldBzrDirFormat()
        self.addCleanup(bzr.BzrProber.formats.remove, old_format.get_format_string())
        bzr.BzrProber.formats.register(old_format.get_format_string(), old_format)
        self.addCleanup(
            controldir.ControlDirFormat._set_default_format,
            controldir.ControlDirFormat.get_default_format(),
        )

        # setup an old format branch we can upgrade from.
        path = "old_format_branch"
        self.make_branch_and_tree(path, format=old_format)
        transport = self.get_transport(path)
        url = transport.base
        display_url = transport.local_abspath(".")
        # check --format takes effect
        controldir.ControlDirFormat._set_default_format(old_format)
        backup_dir = "backup.bzr.~1~"
        (out, err) = self.run_bzr(["upgrade", "--format=2a", url])
        self.assertEqualDiff(
            """Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting upgrade from old test format to 2a
finished
""".format(display_url, display_url, display_url, display_url, backup_dir),
            out,
        )
        self.assertEqualDiff("", err)
        self.assertTrue(
            isinstance(
                controldir.ControlDir.open(self.get_url(path))._format,
                bzrdir.BzrDirMetaFormat1,
            )
        )

    def test_upgrade_explicit_knit(self):
        # users can force an upgrade to knit format from a metadir pack 0.92
        # branch to a 2a branch.
        self.make_branch_and_tree("branch", format="knit")
        transport = self.get_transport("branch")
        url = transport.base
        display_url = transport.local_abspath(".")
        # check --format takes effect
        backup_dir = "backup.bzr.~1~"
        (out, err) = self.run_bzr(["upgrade", "--format=pack-0.92", url])
        self.assertEqualDiff(
            """Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir),
            out,
        )
        self.assertEqualDiff("", err)
        converted_dir = controldir.ControlDir.open(self.get_url("branch"))
        self.assertTrue(isinstance(converted_dir._format, bzrdir.BzrDirMetaFormat1))
        self.assertTrue(
            isinstance(
                converted_dir.open_repository()._format, RepositoryFormatKnitPack1
            )
        )

    def test_upgrade_repo(self):
        self.run_bzr("init-shared-repository --format=pack-0.92 repo")
        self.run_bzr("upgrade --format=2a repo")

    def assertLegalOption(self, option_str):
        # Confirm that an option is legal. (Lower level tests are
        # expected to validate the actual functionality.)
        self.run_bzr("init --format=pack-0.92 branch-foo")
        self.run_bzr("upgrade --format=2a branch-foo {}".format(option_str))

    def assertBranchFormat(self, dir, format):
        branch = controldir.ControlDir.open_tree_or_branch(self.get_url(dir))[1]
        branch_format = branch._format
        meta_format = controldir.format_registry.make_controldir(format)
        expected_format = meta_format.get_branch_format()
        self.assertEqual(expected_format, branch_format)

    def test_upgrade_clean_supported(self):
        self.assertLegalOption("--clean")
        self.assertBranchFormat("branch-foo", "2a")
        backup_bzr_dir = os.path.join("branch-foo", "backup.bzr.~1~")
        self.assertFalse(os.path.exists(backup_bzr_dir))

    def test_upgrade_dry_run_supported(self):
        self.assertLegalOption("--dry-run")
        self.assertBranchFormat("branch-foo", "pack-0.92")

    def test_upgrade_permission_check(self):
        """'backup.bzr' should retain permissions of .bzr. Bug #262450."""
        self.requireFeature(features.posix_permissions_feature)
        old_perms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        backup_dir = "backup.bzr.~1~"
        self.run_bzr("init --format=1.6")
        os.chmod(".bzr", old_perms)
        self.run_bzr("upgrade")
        new_perms = os.stat(backup_dir).st_mode & 0o777
        self.assertTrue(new_perms == old_perms)

    def test_upgrade_with_existing_backup_dir(self):
        self.make_branch_and_tree("old_format_branch", format="knit")
        t = self.get_transport("old_format_branch")
        url = t.base
        display_url = t.local_abspath(".")
        backup_dir1 = "backup.bzr.~1~"
        backup_dir2 = "backup.bzr.~2~"
        # explicitly create backup_dir1. brz should create the .~2~ directory
        # as backup
        t.mkdir(backup_dir1)
        (out, err) = self.run_bzr(["upgrade", "--format=2a", url])
        self.assertEqualDiff(
            """Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir2),
            out,
        )
        self.assertEqualDiff("", err)
        self.assertTrue(
            isinstance(
                controldir.ControlDir.open(self.get_url("old_format_branch"))._format,
                bzrdir.BzrDirMetaFormat1,
            )
        )
        self.assertTrue(t.has(backup_dir2))


class SFTPTests(TestCaseWithSFTPServer):
    """Tests for upgrade over sftp."""

    def test_upgrade_url(self):
        self.run_bzr("init --format=pack-0.92")
        t = self.get_transport()
        url = t.base
        display_url = urlutils.unescape_for_display(url, "utf-8")
        out, err = self.run_bzr(["upgrade", "--format=2a", url])
        backup_dir = "backup.bzr.~1~"
        self.assertEqualDiff(
            """Upgrading branch {} ...
starting upgrade of {}
making backup of {}.bzr
  to {}{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir),
            out,
        )
        self.assertEqual("", err)


class UpgradeRecommendedTests(TestCaseWithTransport):
    def test_recommend_upgrade_wt4(self):
        # using a deprecated format gives a warning
        self.run_bzr("init --format=knit a")
        out, err = self.run_bzr("status a")
        self.assertContainsRe(err, "brz upgrade .*[/\\\\]a")

    def test_no_upgrade_recommendation_from_bzrdir(self):
        # we should only get a recommendation to upgrade when we're accessing
        # the actual workingtree, not when we only open a bzrdir that contains
        # an old workngtree
        self.run_bzr("init --format=knit a")
        out, err = self.run_bzr("revno a")
        if err.find("upgrade") > -1:
            self.fail("message shouldn't suggest upgrade:\n{}".format(err))

    def test_upgrade_shared_repo(self):
        self.make_repository("repo", format="2a", shared=True)
        self.make_branch_and_tree("repo/branch", format="pack-0.92")
        self.get_transport("repo/branch/.bzr/repository").delete_tree(".")
        out, err = self.run_bzr(["upgrade"], working_dir="repo/branch")
